"""
.. module:: samplers.mcmc

:Synopsis: Blocked fast-slow Metropolis sampler (Lewis 1304.4473)
:Author: Antony Lewis (for the CosmoMC sampler, wrapped for cobaya by Jesus Torrado)
"""

# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division
import six

# Global
import os
from copy import deepcopy
from itertools import chain
import numpy as np
from collections import OrderedDict as odict
import logging
from pandas import DataFrame

# Local
from cobaya.sampler import Sampler
from cobaya.mpi import get_mpi_size, get_mpi_rank, get_mpi_comm
from cobaya.mpi import more_than_one_process, am_single_or_primary_process, sync_processes
from cobaya.collection import Collection, OnePoint
from cobaya.conventions import _weight, _p_proposal, _p_renames, _sampler, _minuslogpost
from cobaya.conventions import _line_width, _path_install
from cobaya.samplers.mcmc.proposal import BlockedProposer
from cobaya.log import HandledException
from cobaya.tools import get_external_function, read_dnumber, relative_to_int
from cobaya.yaml import yaml_dump_file
from cobaya.output import OutputDummy

ignore_at_resume = ["burn_in", "callback_function", "callback_every", "max_tries",
                    "check_every", "output_every", "learn_proposal_Rminus1_max",
                    "learn_proposal_Rminus1_max_early", "learn_proposal_Rminus1_min",
                    "max_samples", "Rminus1_stop", "Rminus1_cl_stop",
                    "Rminus1_cl_level", "covmat", "covmat_params", "proposal_scale",
                    "Rminus1_last", "converged"]


class mcmc(Sampler):

    def initialize(self):
        """Initializes the sampler:
        creates the proposal distribution and draws the initial sample."""
        self.log.debug("Initializing")
        for p in [
            "burn_in", "max_tries", "output_every", "check_every", "callback_every"]:
            setattr(
                self, p, read_dnumber(getattr(self, p), self.model.prior.d(), dtype=int))
        if self.callback_every is None:
            self.callback_every = self.check_every
        # Burning-in countdown -- the +1 accounts for the initial point (always accepted)
        self.burn_in_left = self.burn_in + 1
        # Max # checkpoints to wait, in case one process dies without sending MPI_ABORT
        self.been_waiting = 0
        self.max_waiting = max(50, self.max_tries / self.model.prior.d())
        if am_single_or_primary_process():
            if self.resuming and (max(self.mpi_size or 0, 1) != max(get_mpi_size(), 1)):
                self.log.error(
                    "Cannot resume a sample with a different number of chains: "
                    "was %d and now is %d.", max(self.mpi_size, 1), max(get_mpi_size(), 1))
                raise HandledException
        sync_processes()
        if not self.resuming and self.output:
            # Delete previous files (if not "forced", the run would have already failed)
            if ((os.path.abspath(self.covmat_filename()) !=
                 os.path.abspath(str(self.covmat)))):
                try:
                    os.remove(self.covmat_filename())
                except OSError:
                    pass
            # There may be more that chains than expected,
            # if #ranks was bigger in a previous run
            i = 0
            while True:
                i += 1
                collection_filename, _ = self.output.prepare_collection(str(i))
                try:
                    os.remove(collection_filename)
                except OSError:
                    break
        # One collection per MPI process: `name` is the MPI rank + 1
        name = str(1 + (lambda r: r if r is not None else 0)(get_mpi_rank()))
        self.collection = Collection(
            self.model, self.output, name=name, resuming=self.resuming)
        self.current_point = OnePoint(
            self.model, OutputDummy({}), name=name)
        # Use standard MH steps by default
        self.get_new_sample = self.get_new_sample_metropolis
        # Prepare oversampling / dragging if applicable
        self.effective_max_samples = self.max_samples
        if self.oversample and self.drag:
            self.log.error("Choose either oversampling or dragging, not both.")
            raise HandledException
        if self.blocking:
            speeds, blocks = self.model.likelihood._check_speeds_of_params(self.blocking)
            if self.oversample or self.drag:
                speeds = relative_to_int(speeds)
        else:
            speeds, blocks = self.model.likelihood._speeds_of_params(
                int_speeds=self.oversample or self.drag, fast_slow=self.drag)
        if self.oversample:
            self.oversampling_factors = speeds
            self.log.info("Oversampling with factors:\n" + "\n".join([
                "   %d : %r" % (f, b) for f, b in zip(self.oversampling_factors, blocks)]))
            self.i_last_slow_block = None
            # No way right now to separate slow and fast
            slow_params = list(self.model.parameterization.sampled_params())
        elif self.drag:
            # For now, no blocking inside either fast or slow: just 2 blocks
            self.i_last_slow_block = 0
            if np.all(speeds == speeds[0]):
                self.log.error("All speeds are equal or too similar: cannot drag! "
                               "Make sure to define accurate likelihoods' speeds.")
                raise HandledException
            # Make the 1st factor 1:
            speeds = [1, speeds[1] / speeds[0]]
            # Target: dragging step taking as long as slow step
            self.drag_interp_steps = self.drag * speeds[1]
            # Per dragging step, the (fast) posterior is evaluated *twice*,
            self.drag_interp_steps /= 2
            self.drag_interp_steps = int(np.round(self.drag_interp_steps))
            fast_params = list(chain(*blocks[1 + self.i_last_slow_block:]))
            # Not too much or too little dragging
            drag_limits = [(int(l) * len(fast_params) if l is not None else l)
                           for l in self.drag_limits]
            if drag_limits[0] is not None and self.drag_interp_steps < drag_limits[0]:
                self.log.warning("Number of dragging steps clipped from below: was not "
                                 "enough to efficiently explore the fast directions -- "
                                 "avoid this limit by decreasing 'drag_limits[0]'.")
                self.drag_interp_steps = drag_limits[0]
            if drag_limits[1] is not None and self.drag_interp_steps > drag_limits[1]:
                self.log.warning("Number of dragging steps clipped from above: "
                                 "excessive, probably inefficient, exploration of the "
                                 "fast directions -- "
                                 "avoid this limit by increasing 'drag_limits[1]'.")
                self.drag_interp_steps = drag_limits[1]
            # Re-scale steps between checkpoint and callback to the slow dimensions only
            slow_params = list(chain(*blocks[:1 + self.i_last_slow_block]))
            self.n_slow = len(slow_params)
            for p in ["check_every", "callback_every"]:
                setattr(self, p, int(getattr(self, p) * self.n_slow / self.model.prior.d()))
            self.log.info(
                "Dragging with oversampling per step:\n" +
                "\n".join(["   %d : %r" % (f, b)
                           for f, b in zip([1, self.drag_interp_steps],
                                           [blocks[0], fast_params])]))
            self.get_new_sample = self.get_new_sample_dragging
        else:
            self.oversampling_factors = [1 for b in blocks]
            slow_params = list(self.model.parameterization.sampled_params())
            self.n_slow = len(slow_params)
        # Turn parameter names into indices
        self.blocks = [
            [list(self.model.parameterization.sampled_params()).index(p) for p in b]
            for b in blocks]
        self.proposer = BlockedProposer(
            self.blocks, oversampling_factors=self.oversampling_factors,
            i_last_slow_block=self.i_last_slow_block, proposal_scale=self.proposal_scale)
        # Build the initial covariance matrix of the proposal, or load from checkpoint
        if self.resuming:
            covmat = np.loadtxt(self.covmat_filename())
            self.log.info("Covariance matrix from checkpoint.")
        else:
            covmat = self.initial_proposal_covmat(slow_params=slow_params)
            self.log.info("Initial covariance matrix.")
        self.log.debug(
            "Sampling with covmat:\n%s",
            DataFrame(covmat, columns=self.model.parameterization.sampled_params(),
                      index=self.model.parameterization.sampled_params()).to_string(
                line_width=_line_width))
        self.proposer.set_covariance(covmat)
        # Prepare callback function
        if self.callback_function is not None:
            self.callback_function_callable = (
                get_external_function(self.callback_function))
        # Useful for getting last points added inside callback function
        self.last_point_callback = 0

    def initial_proposal_covmat(self, slow_params=None):
        """
        Build the initial covariance matrix, using the data provided, in descending order
        of priority:
        1. "covmat" field in the "mcmc" sampler block.
        2. "proposal" field for each parameter.
        3. variance of the reference pdf.
        4. variance of the prior pdf.

        The covariances between parameters when both are present in a covariance matrix
        provided through option 1 are preserved. All other covariances are assumed 0.
        """
        params_infos = self.model.parameterization.sampled_params_info()
        covmat = np.diag([np.nan] * len(params_infos))
        # Try to generate it automatically
        if isinstance(self.covmat, six.string_types) and self.covmat.lower() == "auto":
            slow_params_info = {
                p: info for p, info in params_infos.items() if p in slow_params}
            auto_covmat = self.model.likelihood._get_auto_covmat(slow_params_info)
            if auto_covmat:
                self.covmat = os.path.join(auto_covmat["folder"], auto_covmat["name"])
                self.log.info("Covariance matrix selected automatically: %s", self.covmat)
            else:
                self.covmat = None
                self.log.info("Could not automatically find a good covmat. "
                              "Will generate from parameter info (proposal and prior).")
        # If given, load and test the covariance matrix
        if isinstance(self.covmat, six.string_types):
            covmat_pre = "{%s}" % _path_install
            if self.covmat.startswith(covmat_pre):
                self.covmat = self.covmat.format(
                    **{_path_install: self.path_install}).replace("/", os.sep)
            try:
                with open(self.covmat, "r") as file_covmat:
                    header = file_covmat.readline()
                loaded_covmat = np.loadtxt(self.covmat)
            except TypeError:
                self.log.error("The property 'covmat' must be a file name,"
                               "but it's '%s'.", str(self.covmat))
                raise HandledException
            except IOError:
                self.log.error("Can't open covmat file '%s'.", self.covmat)
                raise HandledException
            if header[0] != "#":
                self.log.error(
                    "The first line of the covmat file '%s' "
                    "must be one list of parameter names separated by spaces "
                    "and staring with '#', and the rest must be a square matrix, "
                    "with one row per line.", self.covmat)
                raise HandledException
            loaded_params = header.strip("#").strip().split()
        elif hasattr(self.covmat, "__getitem__"):
            if not self.covmat_params:
                self.log.error(
                    "If a covariance matrix is passed as a numpy array, "
                    "you also need to pass the parameters it corresponds to "
                    "via 'covmat_params: [name1, name2, ...]'.")
                raise HandledException
            loaded_params = self.covmat_params
            loaded_covmat = np.array(self.covmat)
        if self.covmat is not None:
            if len(loaded_params) != len(set(loaded_params)):
                self.log.error(
                    "There are duplicated parameters in the header of the "
                    "covmat file '%s' ", self.covmat)
                raise HandledException
            if len(loaded_params) != loaded_covmat.shape[0]:
                self.log.error(
                    "The number of parameters in the header of '%s' and the "
                    "dimensions of the matrix do not coincide.", self.covmat)
                raise HandledException
            if not (np.allclose(loaded_covmat.T, loaded_covmat) and
                    np.all(np.linalg.eigvals(loaded_covmat) > 0)):
                self.log.error(
                    "The covmat loaded from '%s' is not a positive-definite, "
                    "symmetric square matrix.", self.covmat)
                raise HandledException
            # Fill with parameters in the loaded covmat
            renames = [[p] + np.atleast_1d(v.get(_p_renames, [])).tolist()
                       for p, v in params_infos.items()]
            renames = odict([[a[0], a] for a in renames])
            indices_used, indices_sampler = zip(*[
                [loaded_params.index(p),
                 [list(params_infos).index(q) for q, a in renames.items() if p in a]]
                for p in loaded_params])
            if not any(indices_sampler):
                self.log.error(
                    "A proposal covariance matrix has been loaded, but none of its "
                    "parameters are actually sampled here. Maybe a mismatch between"
                    " parameter names in the covariance matrix and the input file?")
                raise HandledException
            indices_used, indices_sampler = zip(*[
                [i, j] for i, j in zip(indices_used, indices_sampler) if j])
            if any(len(j) - 1 for j in indices_sampler):
                first = next(j for j in indices_sampler if len(j) > 1)
                self.log.error(
                    "The parameters %s have duplicated aliases. Can't assign them an "
                    "element of the covariance matrix unambiguously.",
                    ", ".join([list(params_infos)[i] for i in first]))
                raise HandledException
            indices_sampler = list(chain(*indices_sampler))
            covmat[np.ix_(indices_sampler, indices_sampler)] = (
                loaded_covmat[np.ix_(indices_used, indices_used)])
            self.log.info(
                "Covariance matrix loaded for params %r",
                [list(params_infos)[i] for i in indices_sampler])
            missing_params = set(params_infos).difference(
                set([list(params_infos)[i] for i in indices_sampler]))
            if missing_params:
                self.log.info(
                    "Missing proposal covariance for params %r",
                    [p for p in self.model.parameterization.sampled_params()
                     if p in missing_params])
            else:
                self.log.info("All parameters' covariance loaded from given covmat.")
        # Fill gaps with "proposal" property, if present, otherwise ref (or prior)
        where_nan = np.isnan(covmat.diagonal())
        if np.any(where_nan):
            covmat[where_nan, where_nan] = np.array(
                [info.get(_p_proposal, np.nan) ** 2
                 for info in params_infos.values()])[where_nan]
            # we want to start learning the covmat earlier
            self.log.info("Covariance matrix " +
                          ("not present" if np.all(where_nan) else "not complete") + ". "
                                                                                     "We will start learning the covariance of the proposal earlier:"
                                                                                     " R-1 = %g (was %g).",
                          self.learn_proposal_Rminus1_max_early,
                          self.learn_proposal_Rminus1_max)
            self.learn_proposal_Rminus1_max = self.learn_proposal_Rminus1_max_early
        where_nan = np.isnan(covmat.diagonal())
        if np.any(where_nan):
            covmat[where_nan, where_nan] = (
                self.model.prior.reference_covmat().diagonal()[where_nan])
        assert not np.any(np.isnan(covmat))
        return covmat

    def run(self):
        """
        Runs the sampler.
        """
        # Get first point, to be discarded -- not possible to determine its weight
        # Still, we need to compute derived parameters, since, as the proposal "blocked",
        # we may be saving the initial state of some block.
        # NB: if resuming but nothing was written (burn-in not finished): re-start
        self.log.info("Initial point:")
        if self.resuming and self.collection.n():
            initial_point = (self.collection[self.collection.sampled_params]
                .ix[self.collection.n() - 1]).values.copy()
            logpost = -(self.collection[_minuslogpost]
                        .ix[self.collection.n() - 1].copy())
            logpriors = -(self.collection[self.collection.minuslogprior_names]
                          .ix[self.collection.n() - 1].copy())
            loglikes = -0.5 * (self.collection[self.collection.chi2_names]
                               .ix[self.collection.n() - 1].copy())
            derived = (self.collection[self.collection.derived_params]
                       .ix[self.collection.n() - 1].values.copy())
        else:
            initial_point = self.model.prior.reference(max_tries=self.max_tries)
            logpost, logpriors, loglikes, derived = self.model.logposterior(initial_point)
        self.current_point.add(initial_point, derived=derived, logpost=logpost,
                               logpriors=logpriors, loglikes=loglikes)
        self.log.info("\n%s", self.current_point.data.to_string(
            index=False, line_width=_line_width))
        # Initial dummy checkpoint (needed when 1st checkpoint not reached in prev. run)
        self.write_checkpoint()
        # Main loop!
        self.log.info("Sampling!" + (
            " (NB: nothing will be printed until %d burn-in samples " % self.burn_in +
            "have been obtained)" if self.burn_in else ""))
        while self.n() < self.effective_max_samples and not self.converged:
            self.get_new_sample()
            # Callback function
            if (hasattr(self, "callback_function_callable") and
                    not (max(self.n(), 1) % self.callback_every) and
                    self.current_point[_weight] == 1):
                self.callback_function_callable(self)
                self.last_point_callback = self.collection.n()
            # Checking convergence and (optionally) learning the covmat of the proposal
            if self.check_all_ready():
                self.check_convergence_and_learn_proposal()
            if self.n() == self.effective_max_samples:
                self.log.info("Reached maximum number of accepted steps allowed. "
                              "Stopping.")
        # Make sure the last batch of samples ( < output_every ) are written
        self.collection._out_update()
        if more_than_one_process():
            Ns = (lambda x: np.array(get_mpi_comm().gather(x)))(self.n())
        else:
            Ns = [self.n()]
        if am_single_or_primary_process():
            self.log.info("Sampling complete after %d accepted steps.", sum(Ns))

    def n(self, burn_in=False):
        """
        Returns the total number of steps taken, including or not burn-in steps depending
        on the value of the `burn_in` keyword.
        """
        return self.collection.n() + (
            0 if not burn_in else self.burn_in - self.burn_in_left + 1)

    def get_new_sample_metropolis(self):
        """
        Draws a new trial point from the proposal pdf and checks whether it is accepted:
        if it is accepted, it saves the old one into the collection and sets the new one
        as the current state; if it is rejected increases the weight of the current state
        by 1.

        Returns:
           ``True`` for an accepted step, ``False`` for a rejected one.
        """
        trial = deepcopy(self.current_point[self.model.parameterization._sampled])
        self.proposer.get_proposal(trial)
        logpost_trial, logprior_trial, loglikes_trial, derived = self.model.logposterior(trial)
        accept = self.metropolis_accept(logpost_trial,
                                        -self.current_point["minuslogpost"])
        self.process_accept_or_reject(accept, trial, derived,
                                      logpost_trial, logprior_trial, loglikes_trial)
        return accept

    def get_new_sample_dragging(self):
        """
        Draws a new trial point in the slow subspace, and gets the corresponding trial
        in the fast subspace by "dragging" the fast parameters.
        Finally, checks the acceptance of the total step using the "dragging" pdf:
        if it is accepted, it saves the old one into the collection and sets the new one
        as the current state; if it is rejected increases the weight of the current state
        by 1.

        Returns:
           ``True`` for an accepted step, ``False`` for a rejected one.
        """
        # Prepare starting and ending points *in the SLOW subspace*
        # "start_" and "end_" mean here the extremes in the SLOW subspace
        start_slow_point = self.current_point[self.model.parameterization._sampled]
        start_slow_logpost = -self.current_point["minuslogpost"]
        end_slow_point = deepcopy(start_slow_point)
        self.proposer.get_proposal_slow(end_slow_point)
        self.log.debug("Proposed slow end-point: %r", end_slow_point)
        # Save derived parameters of delta_slow jump, in case I reject all the dragging
        # steps but accept the move in the slow direction only
        end_slow_logpost, end_slow_logprior, end_slow_loglikes, derived = (
            self.model.logposterior(end_slow_point))
        if end_slow_logpost == -np.inf:
            self.current_point.increase_weight(1)
            return False
        # trackers of the dragging
        current_start_point = start_slow_point
        current_end_point = end_slow_point
        current_start_logpost = start_slow_logpost
        current_end_logpost = end_slow_logpost
        current_end_logprior = end_slow_logprior
        current_end_loglikes = end_slow_loglikes
        # accumulators for the "dragging" probabilities to be metropolist-tested
        # at the end of the interpolation
        start_drag_logpost_acc = start_slow_logpost
        end_drag_logpost_acc = end_slow_logpost
        # start dragging
        for i_step in range(1, 1 + self.drag_interp_steps):
            self.log.debug("Dragging step: %d", i_step)
            # take a step in the fast direction in both slow extremes
            delta_fast = np.zeros(len(current_start_point))
            self.proposer.get_proposal_fast(delta_fast)
            self.log.debug("Proposed fast step delta: %r", delta_fast)
            proposal_start_point = deepcopy(current_start_point)
            proposal_start_point += delta_fast
            proposal_end_point = deepcopy(current_end_point)
            proposal_end_point += delta_fast
            # get the new extremes for the interpolated probability
            # (reject if any of them = -inf; avoid evaluating both if just one fails)
            # Force the computation of the (slow blocks) derived params at the starting
            # point, but discard them, since they contain the starting point's fast ones,
            # not used later -- save the end point's ones.
            proposal_start_logpost = self.model.logposterior(proposal_start_point)[0]
            proposal_end_logpost, proposal_end_logprior, \
            proposal_end_loglikes, derived_proposal_end = (
                self.model.logposterior(proposal_end_point)
                if proposal_start_logpost > -np.inf
                else (-np.inf, None, [], []))
            if proposal_start_logpost > -np.inf and proposal_end_logpost > -np.inf:
                # create the interpolated probability and do a Metropolis test
                frac = i_step / (1 + self.drag_interp_steps)
                proposal_interp_logpost = ((1 - frac) * proposal_start_logpost
                                           + frac * proposal_end_logpost)
                current_interp_logpost = ((1 - frac) * current_start_logpost
                                          + frac * current_end_logpost)
                accept_drag = self.metropolis_accept(proposal_interp_logpost,
                                                     current_interp_logpost)
            else:
                accept_drag = False
            self.log.debug("Dragging step: %s", ("accepted" if accept_drag else "rejected"))
            # If the dragging step was accepted, do the drag
            if accept_drag:
                current_start_point = proposal_start_point
                current_start_logpost = proposal_start_logpost
                current_end_point = proposal_end_point
                current_end_logpost = proposal_end_logpost
                current_end_logprior = proposal_end_logprior
                current_end_loglikes = proposal_end_loglikes
                derived = derived_proposal_end
            # In any case, update the dragging probability for the final metropolis test
            start_drag_logpost_acc += current_start_logpost
            end_drag_logpost_acc += current_end_logpost
        # Test for the TOTAL step
        accept = self.metropolis_accept(end_drag_logpost_acc / self.drag_interp_steps,
                                        start_drag_logpost_acc / self.drag_interp_steps)
        self.process_accept_or_reject(
            accept, current_end_point, derived,
            current_end_logpost, current_end_logprior, current_end_loglikes)
        self.log.debug("TOTAL step: %s", ("accepted" if accept else "rejected"))
        return accept

    def metropolis_accept(self, logp_trial, logp_current):
        """
        Symmetric-proposal Metropolis-Hastings test.

        Returns:
           ``True`` or ``False``.
        """
        if logp_trial == -np.inf:
            return False
        elif logp_trial > logp_current:
            return True
        else:
            return np.random.exponential() > (logp_current - logp_trial)

    def process_accept_or_reject(self, accept_state, trial=None, derived=None,
                                 logpost_trial=None, logprior_trial=None, loglikes_trial=None):
        """Processes the acceptance/rejection of the new point."""
        if accept_state:
            # add the old point to the collection (if not burning or initial point)
            if self.burn_in_left <= 0:
                self.current_point.add_to_collection(self.collection)
                self.log.debug("New sample, #%d: \n   %r", self.n(), self.current_point)
                if self.n() % self.output_every == 0:
                    self.collection._out_update()
            else:
                self.burn_in_left -= 1
                self.log.debug("Burn-in sample:\n   %r", self.current_point)
                if self.burn_in_left == 0 and self.burn_in:
                    self.log.info("Finished burn-in phase: discarded %d accepted steps.",
                                  self.burn_in)
            # set the new point as the current one, with weight one
            self.current_point.add(trial, derived=derived, weight=1, logpost=logpost_trial,
                                   logpriors=logprior_trial, loglikes=loglikes_trial)
        else:  # not accepted
            self.current_point.increase_weight(1)
            # Failure criterion: chain stuck! (but be more permissive during burn_in)
            max_tries_now = self.max_tries * (1 + (10 - 1) * np.sign(self.burn_in_left))
            if self.current_point[_weight] > max_tries_now:
                self.collection._out_update()
                self.log.error(
                    "The chain has been stuck for %d attempts. Stopping sampling. "
                    "If this has happened often, try improving your "
                    "reference point/distribution. Alternatively (though not advisable) "
                    "make 'max_tries: np.inf' (or 'max_tries: .inf' in yaml)",
                    max_tries_now)
                raise HandledException

    # Functions to check convergence and learn the covariance of the proposal distribution

    def check_all_ready(self):
        """
        Checks if the chain(s) is(/are) ready to check convergence and, if requested,
        learn a new covariance matrix for the proposal distribution.
        """
        msg_ready = ("Ready to check convergence" +
                     (" and learn a new proposal covmat" if self.learn_proposal else ""))
        # If *just* (weight==1) got ready to check+learn
        if (self.n() > 0 and self.current_point[_weight] == 1 and
                not (self.n() % self.check_every)):
            self.log.info("Checkpoint: %d samples accepted.", self.n())
            if more_than_one_process():
                self.been_waiting += 1
                if self.been_waiting > self.max_waiting:
                    self.log.error(
                        "Waiting for too long for all chains to be ready. "
                        "Maybe one of them is stuck or died unexpectedly?")
                    raise HandledException
            self.model.dump_timing()
            # If not MPI size > 1, we are ready
            if not more_than_one_process():
                if msg_ready:
                    self.log.info(msg_ready)
                return True
            # If MPI, tell the rest that we are ready -- we use a "gather"
            # ("reduce" was problematic), but we are in practice just pinging
            if not hasattr(self, "req"):  # just once!
                self.all_ready = np.empty(get_mpi_size())
                self.req = get_mpi_comm().Iallgather(
                    np.array([1.]), self.all_ready)
                self.log.info(msg_ready + " (waiting for the rest...)")
        # If all processes are ready to learn (= communication finished)
        if self.req.Test() if hasattr(self, "req") else False:
            # Sanity check: actually all processes have finished
            assert np.all(self.all_ready == 1), (
                "This should not happen! Notify the developers. (Got %r)", self.all_ready)
            if more_than_one_process() and am_single_or_primary_process():
                self.log.info("All chains are r" + msg_ready[1:])
            delattr(self, "req")
            self.been_waiting = 0
            # Just in case, a barrier here
            sync_processes()
            return True
        return False

    def check_convergence_and_learn_proposal(self):
        """
        Checks the convergence of the sampling process (MPI only), and, if requested,
        learns a new covariance matrix for the proposal distribution from the covariance
        of the last samples.
        """
        if more_than_one_process():
            # Compute and gather means, covs and CL intervals of last half of chains
            mean = self.collection.mean(first=int(self.n() / 2))
            cov = self.collection.cov(first=int(self.n() / 2))
            mcsamples = self.collection._sampled_to_getdist_mcsamples(first=int(self.n() / 2))
            try:
                bound = np.array([[
                    mcsamples.confidence(i, limfrac=self.Rminus1_cl_level / 2., upper=which)
                    for i in range(self.model.prior.d())] for which in [False, True]]).T
                success_bounds = True
            except:
                bound = None
                success_bounds = False
            Ns, means, covs, bounds = map(
                lambda x: np.array(get_mpi_comm().gather(x)),
                [self.n(), mean, cov, bound])
        else:
            # Compute and gather means, covs and CL intervals of last m-1 chain fractions
            m = 1 + self.Rminus1_single_split
            cut = int(self.collection.n() / m)
            if cut <= 1:
                self.log.error("Not enough points in chain to check convergence. "
                               "Increase `check_every` or reduce `Rminus1_single_split`.")
                raise HandledException
            Ns = (m - 1) * [cut]
            means = np.array(
                [self.collection.mean(first=i * cut, last=(i + 1) * cut - 1) for i in range(1, m)])
            covs = np.array(
                [self.collection.cov(first=i * cut, last=(i + 1) * cut - 1) for i in range(1, m)])
            # No logging of warnings temporarily, so getdist won't complain unnecessarily
            logging.disable(logging.WARNING)
            mcsampleses = [
                self.collection._sampled_to_getdist_mcsamples(
                    first=i * cut, last=(i + 1) * cut - 1)
                for i in range(1, m)]
            logging.disable(logging.NOTSET)
            try:
                bounds = [np.array(
                    [[mcs.confidence(i, limfrac=self.Rminus1_cl_level / 2., upper=which)
                      for i in range(self.model.prior.d())] for which in [False, True]]).T
                          for mcs in mcsampleses]
                success_bounds = True
            except:
                bounds = None
                success_bounds = False
        # Compute convergence diagnostics
        if am_single_or_primary_process():
            # "Within" or "W" term -- our "units" for assessing convergence
            # and our prospective new covariance matrix
            mean_of_covs = np.average(covs, weights=Ns, axis=0)
            # "Between" or "B" term
            # We don't weight with the number of samples in the chains here:
            # shorter chains will likely be outliers, and we want to notice them
            cov_of_means = np.atleast_2d(np.cov(means.T))  # , fweights=Ns)
            # For numerical stability, we turn mean_of_covs into correlation matrix:
            #   rho = (diag(Sigma))^(-1/2) * Sigma * (diag(Sigma))^(-1/2)
            # and apply the same transformation to the mean of covs (same eigenvals!)
            diagSinvsqrt = np.diag(np.power(np.diag(cov_of_means), -0.5))
            corr_of_means = diagSinvsqrt.dot(cov_of_means).dot(diagSinvsqrt)
            norm_mean_of_covs = diagSinvsqrt.dot(mean_of_covs).dot(diagSinvsqrt)
            # Cholesky of (normalized) mean of covs and eigvals of Linv*cov_of_means*L
            try:
                L = np.linalg.cholesky(norm_mean_of_covs)
            except np.linalg.LinAlgError:
                self.log.warning(
                    "Negative covariance eigenvectors. "
                    "This may mean that the covariance of the samples does not "
                    "contain enough information at this point. "
                    "Skipping this checkpoint")
                success = False
            else:
                Linv = np.linalg.inv(L)
                try:
                    eigvals = np.linalg.eigvalsh(Linv.dot(corr_of_means).dot(Linv.T))
                    success = True
                except np.linalg.LinAlgError:
                    self.log.warning("Could not compute eigenvalues. "
                                     "Skipping this checkpoint.")
                    success = False
                if success:
                    Rminus1 = max(np.abs(eigvals))
                    # For real square matrices, a possible def of the cond number is:
                    condition_number = Rminus1 / min(np.abs(eigvals))
                    self.log.debug("Condition number = %g", condition_number)
                    self.log.debug("Eigenvalues = %r", eigvals)
                    self.log.info(
                        "Convergence of means: R-1 = %f after %d accepted steps" % (
                            Rminus1, (sum(Ns) if more_than_one_process() else self.n())) +
                        (" = sum(%r)" % list(Ns) if more_than_one_process() else ""))
                    # Have we converged in means?
                    # (criterion must be fulfilled twice in a row)
                    if max(Rminus1, self.Rminus1_last) < self.Rminus1_stop:
                        # Check the convergence of the bounds of the confidence intervals
                        # Same as R-1, but with the rms deviation from the mean bound
                        # in units of the mean standard deviation of the chains
                        if success_bounds:
                            Rminus1_cl = (np.std(bounds, axis=0).T /
                                          np.sqrt(np.diag(mean_of_covs)))
                            self.log.debug("normalized std's of bounds = %r", Rminus1_cl)
                            self.log.info(
                                "Convergence of bounds: R-1 = %f after %d " % (
                                    np.max(Rminus1_cl),
                                    (sum(Ns) if more_than_one_process() else self.n())) +
                                "accepted steps" +
                                (" = sum(%r)" % list(Ns) if more_than_one_process() else ""))
                            if np.max(Rminus1_cl) < self.Rminus1_cl_stop:
                                self.converged = True
                                self.log.info("The run has converged!")
                            self._Ns = Ns
                        else:
                            self.log.info("Computation of the bounds was not possible. "
                                          "Waiting until the next checkpoint")
        if more_than_one_process():
            # Broadcast and save the convergence status and the last R-1 of means
            success = get_mpi_comm().bcast(
                (success if am_single_or_primary_process() else None), root=0)
            if success:
                self.Rminus1_last = get_mpi_comm().bcast(
                    (Rminus1 if am_single_or_primary_process() else None), root=0)
                self.converged = get_mpi_comm().bcast(
                    (self.converged if am_single_or_primary_process() else None), root=0)
        else:
            if success:
                self.Rminus1_last = Rminus1
        # Do we want to learn a better proposal pdf?
        if self.learn_proposal and not self.converged and success:
            good_Rminus1 = (self.learn_proposal_Rminus1_max >
                            self.Rminus1_last > self.learn_proposal_Rminus1_min)
            if not good_Rminus1:
                if am_single_or_primary_process():
                    self.log.info("Bad convergence statistics: "
                                  "waiting until the next checkpoint.")
                return
            if more_than_one_process():
                if not am_single_or_primary_process():
                    mean_of_covs = np.empty((self.model.prior.d(), self.model.prior.d()))
                get_mpi_comm().Bcast(mean_of_covs, root=0)
            else:
                mean_of_covs = covs[0]
            try:
                self.proposer.set_covariance(mean_of_covs)
            except:
                self.log.debug("Updating covariance matrix failed unexpectedly. "
                               "waiting until next checkpoint.")
            if am_single_or_primary_process():
                self.log.info("Updated covariance matrix of proposal pdf.")
                self.log.debug("%r", mean_of_covs)
        # Save checkpoint info
        self.write_checkpoint()

    def write_checkpoint(self):
        if am_single_or_primary_process() and self.output:
            checkpoint_filename = self.checkpoint_filename()
            covmat_filename = self.covmat_filename()
            np.savetxt(covmat_filename, self.proposer.get_covariance(), header=" ".join(
                list(self.model.parameterization.sampled_params())))
            checkpoint_info = {_sampler: {self.name: odict([
                ["converged", bool(self.converged)],
                ["Rminus1_last", self.Rminus1_last],
                ["proposal_scale", self.proposer.get_scale()],
                ["blocks", self.blocks],
                ["oversampling_factors", self.oversampling_factors],
                ["i_last_slow_block", self.i_last_slow_block],
                ["burn_in", (self.burn_in  # initial: repeat burn-in if not finished
                             if not self.n() and self.burn_in_left else
                             "d")],  # to avoid overweighting last point of prev. run
                ["mpi_size", get_mpi_size()]])}}
            yaml_dump_file(checkpoint_filename, checkpoint_info, error_if_exists=False)
            self.log.debug("Dumped checkpoint info and current covmat.")

    # Finally: returning the computed products ###########################################

    def products(self):
        """
        Auxiliary function to define what should be returned in a scripted call.

        Returns:
           The sample ``Collection`` containing the accepted steps.
        """
        return {"sample": self.collection}



"""
.. module:: install

:Synopsis: Tools and script to install the modules requested in the given input.
:Author: Jesus Torrado

"""
# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

# Global
import os
import sys
import subprocess
import traceback
import logging
from importlib import import_module
import shutil
from six import string_types
from pkg_resources import parse_version

# Local
from cobaya.log import logger_setup, HandledException
from cobaya.tools import get_folder, make_header, warn_deprecation
from cobaya.input import get_modules
from cobaya.conventions import _package, _code, _data, _likelihood, _external
from cobaya.conventions import _modules_path_arg, _path_install

log = logging.getLogger(__name__.split(".")[-1])


def install(*infos, **kwargs):
    if not log.root.handlers:
        logger_setup()
    path = kwargs.get("path", ".")
    if not path:
        # See if we can get one (and only one) from infos
        paths = set([p for p in [info.get(_path_install) for info in infos] if p])
        if len(paths) == 1:
            path = paths[0]
        else:
            print("logging?")
            log.error("No 'path' argument given and could not extract one (and only one) "
                      "from the infos.")
            raise HandledException
    abspath = os.path.abspath(path)
    log.info("Installing modules at '%s'\n", abspath)
    kwargs_install = {"force": kwargs.get("force", False),
                      "no_progress_bars": kwargs.get("no_progress_bars")}
    for what in (_code, _data):
        kwargs_install[what] = kwargs.get(what, True)
        spath = os.path.join(abspath, what)
        if kwargs_install[what] and not os.path.exists(spath):
            try:
                os.makedirs(spath)
            except OSError:
                log.error("Could not create the desired installation folder '%s'", spath)
                raise HandledException
    failed_modules = []
    for kind, modules in get_modules(*infos).items():
        for module in modules:
            print(make_header(kind, module))
            module_folder = get_folder(module, kind, sep=".", absolute=False)
            try:
                imported_module = import_module(module_folder, package=_package)
            except ImportError:
                if kind == _likelihood:
                    info = (next(info for info in infos
                                 if module in info.get(_likelihood, {}))
                            [_likelihood][module]) or {}
                    if isinstance(info, string_types) or _external in info:
                        log.warning("Module '%s' is a custom likelihood. "
                                    "Nothing to do.\n", module)
                        flag = False
                    else:
                        log.error("Module '%s' not recognized.\n" % module)
                        failed_modules += ["%s:%s" % (kind, module)]
                continue
            is_installed = getattr(imported_module, "is_installed", None)
            if is_installed is None:
                log.info("Built-in module: nothing to do.\n")
                continue
            if is_installed(path=abspath, **kwargs_install):
                log.info("External module already installed.\n")
                if kwargs_install["force"]:
                    log.info("Forcing re-installation, as requested.")
                else:
                    log.info("Doing nothing.\n")
                    continue
            try:
                success = imported_module.install(path=abspath, **kwargs_install)
            except:
                traceback.print_exception(*sys.exc_info(), file=sys.stdout)
                log.error("An unknown error occurred. Delete the modules folder and try "
                          "again. Notify the developers if this error persists.")
                success = False
            if success:
                log.info("Successfully installed!\n")
            else:
                log.error("Installation failed! Look at the error messages above. "
                          "Solve them and try again, or, if you are unable to solve, "
                          "install this module manually.")
                failed_modules += ["%s:%s" % (kind, module)]
                continue
            # test installation
            if not is_installed(path=abspath, **kwargs_install):
                log.error("Installation apparently worked, "
                          "but the subsequent installation test failed! "
                          "Look at the error messages above. "
                          "Solve them and try again, or, if you are unable to solve, "
                          "install this module manually.")
                failed_modules += ["%s:%s" % (kind, module)]
    if failed_modules:
        bullet = "\n - "
        log.error("The installation (or installation test) of some module(s) has failed: "
                  "%s\nCheck output of the installer of each module above "
                  "for precise error info.\n",
                  bullet + bullet.join(failed_modules))
        raise HandledException


def download_file(filename, path, no_progress_bars=False, decompress=False, logger=None):
    log = logger or logging.getLogger(__name__)
    try:
        from wget import download, bar_thermometer
        wget_kwargs = {"out": path, "bar":
            (bar_thermometer if not no_progress_bars else None)}
        filename = download(filename, **wget_kwargs)
    except:
        log.error("Error downloading file '%s' to folder '%s'", filename, path)
        return False
    finally:
        print("")
    log.debug('Got: %s' % filename)
    if not decompress:
        return True
    import tarfile
    extension = os.path.splitext(filename)[-1][1:]
    if extension == "tgz":
        extension = "gz"
    try:
        tar = tarfile.open(filename, "r:" + extension)
        tar.extractall(path)
        tar.close()
        os.remove(filename)
        log.debug('Decompressed: %s' % filename)
        return True
    except:
        log.error("Error decompressing downloaded file! Corrupt file?")
        return False


def download_github_release(directory, repo_name, release_name, repo_rename=None,
                            no_progress_bars=False):
    if "/" in repo_name:
        github_user = repo_name[:repo_name.find("/")]
        repo_name = repo_name[repo_name.find("/")+1:]
    else:
        github_user = "CobayaSampler"
    if not os.path.exists(directory):
        os.makedirs(directory)
    filename = (r"https://github.com/" + github_user + "/" + repo_name +
                "/archive/" + release_name + ".tar.gz")
    if not download_file(filename, directory, decompress=True,
                         no_progress_bars=no_progress_bars, logger=log):
        return False
    # Remove version number from directory name
    w_version = next(d for d in os.listdir(directory)
                     if (d.startswith(repo_name) and len(d) != len(repo_name)))
    repo_rename = repo_rename or repo_name
    repo_path = os.path.join(directory, repo_rename)
    if os.path.exists(repo_path):
        shutil.rmtree(repo_path)
    os.rename(os.path.join(directory, w_version), repo_path)
    log.info("%s %s downloaded and decompressed correctly.", repo_name, release_name)
    return True


def pip_install(packages, upgrade=False):
    """
    Takes package name or list of them.

    Uses ``--user`` flag if does not appear to have write permission to python path

    Returns exit status.
    """
    if hasattr(packages, "split"):
        packages = [packages]
    cmd = [sys.executable, '-m', 'pip', 'install']
    if not os.access(os.path.dirname(sys.executable), os.W_OK):
        cmd += ['--user']
    if upgrade:
        cmd += ['--upgrade']
    res = subprocess.call(cmd + packages)
    if res:
        log.error("pip: error installing packages '%s'", packages)
    return res


def check_gcc_version(min_version="6.4", error_returns=None):
    try:
        version = subprocess.check_output(
            "gcc -dumpversion", shell=True, stderr=subprocess.STDOUT).decode().strip()
    except:
        return error_returns
    # Change in gcc >= 7: -dumpversion only dumps major version
    if not "." in version:
        version = subprocess.check_output(
            "gcc -dumpfullversion", shell=True, stderr=subprocess.STDOUT).decode().strip()
    return parse_version(str(min_version)) <= parse_version(version)


# Command-line script ####################################################################

def install_script():
    from cobaya.mpi import am_single_or_primary_process
    if am_single_or_primary_process():
        warn_deprecation()
        # Configure the logger ASAP
        logger_setup()
        log = logging.getLogger(__name__.split(".")[-1])
        # Parse arguments
        import argparse
        parser = argparse.ArgumentParser(
            description="Cobaya's installation tool for external modules.")
        parser.add_argument("files", action="store", nargs="+", metavar="input_file.yaml",
                            help="One or more input files "
                                 "(or simply 'polychord', or 'cosmo' "
                                 "for a basic collection of cosmological modules)")
        parser.add_argument("-" + _modules_path_arg[0], "--" + _modules_path_arg,
                            action="store", nargs=1, required=True,
                            metavar="/install/path",
                            help="Desired path where to install external modules.")
        parser.add_argument("-f", "--force", action="store_true", default=False,
                            help="Force re-installation of apparently installed modules.")
        parser.add_argument("--no-progress-bars", action="store_true", default=False,
                            help="No progress bars shown. Shorter logs (used in Travis).")
        group_just = parser.add_mutually_exclusive_group(required=False)
        group_just.add_argument("-C", "--just-code", action="store_false", default=True,
                                help="Install code of the modules.", dest=_data)
        group_just.add_argument("-D", "--just-data", action="store_false", default=True,
                                help="Install data of the modules.", dest=_code)
        arguments = parser.parse_args()
        if arguments.files == ["cosmo"]:
            log.info("Installing cosmological modules (input files will be ignored)")
            from cobaya.cosmo_input import install_basic
            infos = [install_basic]
        elif arguments.files == ["polychord"]:
            infos = [{"sampler": {"polychord": None}}]
        else:
            from cobaya.input import load_input
            infos = [load_input(f) for f in arguments.files]
        # Launch installer
        install(*infos, path=getattr(arguments, _modules_path_arg)[0],
                **{arg: getattr(arguments, arg)
                   for arg in ["force", _code, _data, "no_progress_bars"]})



# Notes about writing more presets:
# ---------------------------------
# - all parameter names below are PLANCK parameter names. They are substituted by the
#   theory-code-specific ones in `create_input`
# - all presets below, either "atomic" (e.g. primordial, geometry...) or for full runs
#   must be OrderedDict's!
# - don't use extra_args for precision parameters! because if the same precision param
#   is mentioned twice at the same time in different fields with different values, there
#   is no facility to take the max (or min). Instead, codify precision needs in terms of
#   requirements in the .needs method of the cosmo code.

# Python 2/3 compatibility
from __future__ import absolute_import, division, print_function

# Global
from collections import OrderedDict as odict
from copy import deepcopy

# Local
from cobaya.conventions import _theory, _params, _likelihood, _sampler, _prior, _p_derived
from cobaya.conventions import _p_ref, _p_proposal, _p_label, _p_dist, _p_drop, _p_value
from cobaya.conventions import _p_renames, _chi2, _separator

_camb = "camb"
_classy = "classy"
_desc = "desc"
_comment = "note"
_extra_args = "extra_args"
_error_msg = "error_msg"
_none = "(None)"

# Theory codes
theory = odict([[_camb, None], [_classy, None]])

# Primordial perturbations
primordial = odict([
    ["SFSR", {
        _desc: "Adiabatic scalar perturbations, power law spectrum",
        _theory: {_camb: None, _classy: None},
        _params: odict([
            ["logA", odict([
                [_prior, odict([["min", 2], ["max", 4]])],
                [_p_ref, odict([[_p_dist, "norm"], ["loc", 3.1], ["scale", 0.001]])],
                [_p_proposal, 0.001], [_p_label, r"\log(10^{10} A_\mathrm{s})"],
                [_p_drop, True]])],
            ["As", odict([
                [_p_value, "lambda logA: 1e-10*np.exp(logA)"],
                [_p_label, r"A_\mathrm{s}"]])],
            ["ns", odict([
                [_prior, odict([["min", 0.8], ["max", 1.2]])],
                [_p_ref, odict([[_p_dist, "norm"], ["loc", 0.96], ["scale", 0.004]])],
                [_p_proposal, 0.002], [_p_label, r"n_\mathrm{s}"]])]])}]])
primordial.update(odict([
    ["SFSR_run", {
        _desc: "Adiabatic scalar perturbations, power law + running spectrum",
        _theory: {_camb: None, _classy: None},
        _params: odict(
            (list(primordial["SFSR"][_params].items()) +
             [["nrun", odict([
                 [_prior, odict([["min", -1], ["max", 1]])],
                 [_p_ref, odict([[_p_dist, "norm"], ["loc", 0], ["scale", 0.005]])],
                 [_p_proposal, 0.001], [_p_label, r"n_\mathrm{run}"]])]]))}]]))
primordial.update(odict([
    ["SFSR_runrun", {
        _desc: "Adiabatic scalar perturbations, power law + 2nd-order running spectrum",
        _theory: {_camb: None, _classy: None},
        _params: odict(
            (list(primordial["SFSR_run"][_params].items()) +
             [["nrunrun", odict([
                 [_prior, odict([["min", -1], ["max", 1]])],
                 [_p_ref, odict([[_p_dist, "norm"], ["loc", 0], ["scale", 0.002]])],
                 [_p_proposal, 0.001], [_p_label, r"n_\mathrm{run,run}"]])]]))}]]))
primordial.update(odict([
    ["SFSR_t", {
        _desc: "Adiabatic scalar+tensor perturbations, "
               "power law spectrum (inflation consistency)",
        _theory: {_camb: {_extra_args: {"nt": None}},
                  _classy: {_extra_args: {"n_t": "scc", "alpha_t": "scc"}}},
        _params: odict(
            (list(primordial["SFSR"][_params].items()) +
             [["r", odict([
                 [_prior, odict([["min", 0], ["max", 3]])],
                 [_p_ref, odict([[_p_dist, "norm"], ["loc", 0], ["scale", 0.03]])],
                 [_p_proposal, 0.03], [_p_label, r"r_{0.05}"]])]]))}]]))
primordial.update(odict([
    ["SFSR_t_nrun", {
        _desc: "Adiabatic scalar+tensor perturbations, "
               "power law + running spectrum (inflation consistency)",
        _theory: {_camb:
                      {_extra_args: primordial["SFSR_t"][_theory][_camb][_extra_args]},
                  _classy:
                      {_extra_args: primordial["SFSR_t"][_theory][_classy][_extra_args]}},
        _params: odict(
            (list(primordial["SFSR_run"][_params].items()) +
             list(primordial["SFSR_t"][_params].items())))}]]))

# Geometry
geometry = odict([
    ["flat", {
        _desc: "Flat FLRW universe",
        _theory: {_camb: None, _classy: None}}],
    ["omegak", {
        _desc: "FLRW model with varying curvature (prior on Omega_k)",
        _theory: {_camb: None, _classy: None},
        _params: odict([
            ["omegak", odict([
                [_prior, odict([["min", -0.3], ["max", 0.3]])],
                [_p_ref, odict([[_p_dist, "norm"], ["loc", -0.0008], ["scale", 0.001]])],
                [_p_proposal, 0.001], [_p_label, r"\Omega_k"]])]])}], ])

# Hubble parameter constraints
H0_min, H0_max = 40, 100
hubble = odict([
    ["H", {
        _desc: "Hubble parameter",
        _theory: {_camb: None, _classy: None},
        _params: odict([
            ["H0", odict([
                [_prior, odict([["min", H0_min], ["max", H0_max]])],
                [_p_ref, odict([[_p_dist, "norm"], ["loc", 70], ["scale", 2]])],
                [_p_proposal, 2], [_p_label, r"H_0"]])]])}],
    ["sound_horizon_last_scattering", {
        _desc: "Angular size of the sound horizon at last scattering "
               "(approximate, if using CAMB)",
        _theory: {
            _camb: {
                _params: odict([
                    ["theta", odict([
                        [_prior, odict([["min", 0.5], ["max", 10]])],
                        [_p_ref,
                         odict([[_p_dist, "norm"], ["loc", 1.0411], ["scale", 0.0004]])],
                        [_p_proposal, 0.0002], [_p_label, r"100\theta_\mathrm{MC}"],
                        [_p_drop, True]])],
                    ["cosmomc_theta", odict([
                        [_p_value, "lambda theta: 1.e-2*theta"], [_p_derived, False]])],
                    ["H0", {_p_label: r"H_0", "min": H0_min, "max": H0_max}]]),
                _extra_args: odict([["theta_H0_range", [H0_min, H0_max]]])},
            _classy: {
                _params: odict([
                    ["theta_s_1e2", odict([
                        [_prior, odict([["min", 0.5], ["max", 10]])],
                        [_p_ref, odict([
                            [_p_dist, "norm"], ["loc", 1.0418], ["scale", 0.0004]])],
                        [_p_proposal, 0.0002], [_p_label, r"100\theta_\mathrm{s}"],
                        [_p_drop, True]])],
                    ["100*theta_s", odict([
                        [_p_value, "lambda theta_s_1e2: theta_s_1e2"],
                        [_p_derived, False]])],
                    ["H0", {_p_label: r"H_0"}]])}}}]])

# Matter sector (minus light species)
matter = odict([
    ["omegab_h2, omegac_h2", {
        _desc: "Flat prior on Omega*h^2 for barions and cold dark matter",
        _theory: {_camb: None, _classy: None},
        _params: odict([
            ["omegabh2", odict([
                [_prior, odict([["min", 0.005], ["max", 0.1]])],
                [_p_ref, odict([[_p_dist, "norm"], ["loc", 0.0221], ["scale", 0.0001]])],
                [_p_proposal, 0.0001], [_p_label, r"\Omega_\mathrm{b} h^2"]])],
            ["omegach2", odict([
                [_prior, odict([["min", 0.001], ["max", 0.99]])],
                [_p_ref, odict([[_p_dist, "norm"], ["loc", 0.12], ["scale", 0.001]])],
                [_p_proposal, 0.0005], [_p_label, r"\Omega_\mathrm{c} h^2"]])],
            ["omegam", {_p_label: r"\Omega_\mathrm{m}"}]])}],
    ["Omegab, Omegam", {
        _desc: "Flat prior on Omega for barions and total matter",
        _theory: {_camb: None, _classy: None},
        _params: odict([
            ["omegab", odict([
                [_prior, odict([["min", 0.03], ["max", 0.07]])],
                [_p_ref, odict([[_p_dist, "norm"], ["loc", 0.0468], ["scale", 0.004]])],
                [_p_proposal, 0.004], [_p_label, r"\Omega_\mathrm{b}"],
                [_p_drop, True]])],
            ["omegam", odict([
                [_prior, odict([["min", 0.1], ["max", 0.9]])],
                [_p_ref, odict([[_p_dist, "norm"], ["loc", 0.295], ["scale", 0.02]])],
                [_p_proposal, 0.02], [_p_label, r"\Omega_\mathrm{m}"],
                [_p_drop, True]])],
            ["omegabh2", odict([
                [_p_value, "lambda omegab, H0: omegab*(H0/100)**2"],
                [_p_label, r"\Omega_\mathrm{b} h^2"]])],
            ["omegach2", odict([
                [_p_value, ("lambda omegam, omegab, mnu, H0: "
                            "(omegam-omegab)*(H0/100)**2-(mnu*(3.046/3)**0.75)/94.07")],
                [_p_label, r"\Omega_\mathrm{c} h^2"]])]
        ])}]])
for m in matter.values():
    m[_params]["omegamh2"] = odict([
        [_p_derived, "lambda omegam, H0: omegam*(H0/100)**2"],
        [_p_label, r"\Omega_\mathrm{m} h^2"]])

# Neutrinos and other extra matter
neutrinos = odict([
    ["one_heavy_planck", {
        _desc: "Two massless nu and one with m=0.06. Neff=3.046",
        _theory: {
            _camb: {_extra_args: {"num_massive_neutrinos": 1, "nnu": 3.046},
                    _params: odict([["mnu", 0.06]])},
            _classy: {_extra_args: {"N_ncdm": 1, "N_ur": 2.0328},
                      _params: odict([
                          ["m_ncdm", {_p_value: 0.06, _p_renames: "mnu"}]])}}}],
    ["varying_mnu", {
        _desc: "Varying total mass of 3 degenerate nu's, with N_eff=3.046",
        _theory: {
            _camb: {_extra_args: {"num_massive_neutrinos": 3, "nnu": 3.046},
                    _params: odict([
                        ["mnu", odict([
                            [_prior, odict([["min", 0], ["max", 5]])],
                            [_p_ref, odict([
                                [_p_dist, "norm"], ["loc", 0.02], ["scale", 0.1]])],
                            [_p_proposal, 0.03], [_p_label, r"\sum m_\nu"]])]])},
            _classy: {_extra_args: {"N_ncdm": 1, "deg_ncdm": 3, "N_ur": 0.00641},
                      _params: odict([
                          ["m_ncdm", odict([
                              [_prior, odict([["min", 0], ["max", 1.667]])],
                              [_p_ref, odict([
                                  [_p_dist, "norm"], ["loc", 0.0067], ["scale", 0.033]])],
                              [_p_proposal, 0.01], [_p_label, r"m_\nu"]])],
                          ["mnu", odict([[_p_derived, "lambda m_ncdm: 3 * m_ncdm"],
                                         [_p_label, r"\sum m_\nu"]])]])}}}],
    ["varying_Neff", {
        _desc: "Varying Neff with two massless nu and one with m=0.06",
        _theory: {
            _camb: {_extra_args: {"num_massive_neutrinos": 1},
                    _params: odict([
                        ["mnu", 0.06],
                        ["nnu", odict([
                            [_prior, odict([["min", 0.05], ["max", 10]])],
                            [_p_ref, odict([
                                [_p_dist, "norm"], ["loc", 3.046], ["scale", 0.05]])],
                            [_p_proposal, 0.05], [_p_label, r"N_\mathrm{eff}"]])]])},
            _classy: {_extra_args: {"N_ncdm": 1},
                      _params: odict([
                          ["m_ncdm", odict([[_p_value, 0.06], [_p_renames, "mnu"]])],
                          ["N_ur", odict([
                              [_prior, odict([["min", 0.0001], ["max", 9]])],
                              [_p_ref, odict([
                                  [_p_dist, "norm"], ["loc", 2.0328], ["scale", 0.05]])],
                              [_p_proposal, 0.05], [_p_label, r"N_\mathrm{ur}"]])],
                          ["nnu", odict([
                              [_p_derived, "lambda Neff: Neff"],
                              [_p_label, r"N_\mathrm{eff}"]])]])}}}]])
neutrinos.update(odict([
    ["varying_mnu_Neff", {
        _desc: "Varying Neff and total mass of 3 degenerate nu's",
        _theory: {
            _camb: {
                _extra_args: {"num_massive_neutrinos": 3},
                _params: odict([
                    ["mnu", deepcopy(
                        neutrinos["varying_mnu"][_theory]["camb"][_params]["mnu"])],
                    ["nnu", deepcopy(
                        neutrinos["varying_Neff"][_theory]["camb"][_params]["nnu"])]])},
            #            _classy: {
            #                _extra_args: {"N_ncdm": 1, "deg_ncdm": 3},
            #                _params: odict([
            #                    ["m_ncdm", deepcopy(
            #                        neutrinos["varying_mnu"][_theory]["classy"][_params]["m_ncdm"])],
            #                    ["mnu", deepcopy(
            #                        neutrinos["varying_mnu"][_theory]["classy"][_params]["mnu"])],
            #                    ["N_ur", deepcopy(
            #                        neutrinos["varying_Neff"][_theory]["classy"][_params]["N_ur"])],
            #                    ["nnu", deepcopy(
            #                        neutrinos["varying_Neff"][_theory]["classy"][_params]["nnu"])]
            #                ])}
        }}]]))
# neutrinos["varying_mnu_Neff"][_theory][_classy][_params]["N_ur"][_p_ref]["loc"] = 0.00641

#    # ["varying_Neff+1sterile", {
#    #     _desc: "Varying Neff plus 1 sterile neutrino (SM nu's with m=0,0,0.06)",
#    #     _theory: {
#    #         _camb: {_extra_args: odict(
#    #             list(neutrinos["varying_Neff"][_theory][_camb][_extra_args].items()) +
#    #             [["accuracy_level", 1.2]])}},
#    #     _params: odict([
#    #         ["nnu", odict([
#    #             [_prior, odict([["min", 3.046], ["max", 10]])],
#    #             [_p_ref, odict([[_p_dist, "norm"], ["loc", 3.046], ["scale", 0.05]])],
#    #             [_p_proposal, 0.05], [_p_label, r"N_\mathrm{eff}"]])],
#    #         ["meffsterile", odict([
#    #             [_prior, odict([["min", 0], ["max", 3]])#,
#    #             [_p_ref, odict([[_p_dist, "norm"], ["loc", 0.1], ["scale", 0.1]])],
#    #             [_p_proposal, 0.03],
#    #             [_p_label, r"m_{\nu,\mathrm{sterile}}^\mathrm{eff}"]])]])}]

# Dark Energy
dark_energy = odict([
    ["lambda", {
        _desc: "Cosmological constant (w=-1)",
        _theory: {_camb: None, _classy: None},
        _params: odict([
            ["omegal", {_p_label: r"\Omega_\Lambda"}]])}],
    ["de_w", {
        _desc: "Varying constant eq of state",
        _theory: {_camb: None,
                  _classy: {_params: {"Omega_Lambda": 0}}},
        _params: odict([
            ["w", odict([
                [_prior, odict([["min", -3], ["max", -0.333]])],
                [_p_ref, odict([[_p_dist, "norm"], ["loc", -0.99], ["scale", 0.02]])],
                [_p_proposal, 0.02], [_p_label, r"w_\mathrm{DE}"]])]])}],
    ["de_w_wa", {
        _desc: "Varying constant eq of state with w(a) = w0 + (1-a) wa",
        _theory: {_camb: None,
                  _classy: {_params: {"Omega_Lambda": 0}}},
        _params: odict([
            ["w", odict([
                [_prior, odict([["min", -3], ["max", 1]])],
                [_p_ref, odict([[_p_dist, "norm"], ["loc", -0.99], ["scale", 0.02]])],
                [_p_proposal, 0.02], [_p_label, r"w_{0,\mathrm{DE}}"]])],
            ["wa", odict([
                [_prior, odict([["min", -3], ["max", 2]])],
                [_p_ref, odict([[_p_dist, "norm"], ["loc", 0], ["scale", 0.05]])],
                [_p_proposal, 0.05], [_p_label, r"w_{a,\mathrm{DE}}"]])]])}]])

# BBN
bbn_derived_camb = odict([
    ["YpBBN", odict([[_p_label, r"Y_P^\mathrm{BBN}"]])],
    ["DHBBN", odict([[_p_derived, r"lambda DH: 10**5*DH"],
                     [_p_label, r"10^5 \mathrm{D}/\mathrm{H}"]])]])
bbn = odict([
    ["consistency", {
        _desc: "Primordial Helium fraction inferred from BBN consistency",
        _theory: {_camb: {_params: bbn_derived_camb},
                  _classy: None},
        _params: odict([
            ["yheused", {_p_label: r"Y_\mathrm{P}"}]])}],
    ["YHe", {
        _desc: "Varying primordial Helium fraction",
        _theory: {_camb: None,
                  _classy: None},
        _params: odict([
            ["yhe", odict([
                [_prior, odict([["min", 0.1], ["max", 0.5]])],
                [_p_ref, odict([[_p_dist, "norm"], ["loc", 0.245], ["scale", 0.006]])],
                [_p_proposal, 0.006], [_p_label, r"Y_\mathrm{P}"]])]])}], ])

# Reionization
reionization = odict([
    ["std", {
        _desc: "Standard reio, lasting delta_z=0.5",
        _theory: {_camb: None, _classy: None},
        _params: odict([
            ["tau", odict([
                [_prior, odict([["min", 0.01], ["max", 0.8]])],
                [_p_ref, odict([[_p_dist, "norm"], ["loc", 0.09], ["scale", 0.01]])],
                [_p_proposal, 0.005], [_p_label, r"\tau_\mathrm{reio}"]])],
            ["zrei", {_p_label: r"z_\mathrm{re}"}]])}],
    ["gauss_prior", {
        _desc: "Standard reio, lasting delta_z=0.5, gaussian prior around tau=0.07",
        _theory: {_camb: None, _classy: None},
        _params: odict([
            ["tau", odict([
                [_prior, odict([[_p_dist, "norm"], ["loc", 0.07], ["scale", 0.02]])],
                [_p_ref, odict([[_p_dist, "norm"], ["loc", 0.07], ["scale", 0.01]])],
                [_p_proposal, 0.005], [_p_label, r"\tau_\mathrm{reio}"]])],
            ["zrei", {_p_label: r"z_\mathrm{re}"}]])}], ])

# EXPERIMENTS ############################################################################
cmb_precision = {_camb: {"lens_potential_accuracy": 1},
                 _classy: {"non linear": "halofit"}}
cmb_sampler_recommended = {"mcmc": {"drag": True, "proposal_scale": 1.9}}

like_cmb = odict([
    [_none, {}],
    ["planck_2015_lensing", {
        _desc: "Planck 2015 (Polarized CMB + lensing)",
        _comment: None,
        _sampler: cmb_sampler_recommended,
        _theory: {theo: {_extra_args: None}  # cmb_precision[theo]}  # not for now
                  for theo in [_camb, _classy]},
        _likelihood: odict([
            ["planck_2015_lowTEB", None],
            ["planck_2015_plikHM_TTTEEE", None],
            ["planck_2015_lensing", None]])}],
    ["planck_2015_lensing_bk15", {
        _desc: "Planck 2015 (Polarized CMB + lensing) + Bicep/Keck-Array 2015",
        _sampler: cmb_sampler_recommended,
        _theory: {theo: {_extra_args: None}  # cmb_precision[theo]}  # not for now
                  for theo in [_camb, _classy]},
        _likelihood: odict([
            ["planck_2015_lowTEB", None],
            ["planck_2015_plikHM_TTTEEE", None],
            ["planck_2015_lensing", None],
            ["bicep_keck_2015", None]])}],
])
like_cmb["planck_2015_lensing_bk15"][_comment] = like_cmb["planck_2015_lensing"][_comment]
# Add common CMB derived parameters
for m in like_cmb.values():
    # Don't add the derived parameter to the no-CMB case!
    if not m:
        continue
    if _params not in m:
        m[_params] = odict()
    m[_params].update(odict([
        ["sigma8", {_p_label: r"\sigma_8"}],
        ["s8h5", odict([
            [_p_derived, "lambda sigma8, H0: sigma8*(H0*1e-2)**(-0.5)"],
            [_p_label, r"\sigma_8/h^{0.5}"]])],
        ["s8omegamp5", odict([
            [_p_derived, "lambda sigma8, omegam: sigma8*omegam**0.5"],
            [_p_label, r"\sigma_8 \Omega_\mathrm{m}^{0.5}"]])],
        ["s8omegamp25", odict([
            [_p_derived, "lambda sigma8, omegam: sigma8*omegam**0.25"],
            [_p_label, r"\sigma_8 \Omega_\mathrm{m}^{0.25}"]])],
        ["A", odict([
            [_p_derived, "lambda As: 1e9*As"],
            [_p_label, r"10^9 A_\mathrm{s}"]])],
        ["clamp", odict([
            [_p_derived, "lambda As, tau: 1e9*As*np.exp(-2*tau)"],
            [_p_label, r"10^9 A_\mathrm{s} e^{-2\tau}"]])],
        ["age", odict([
            [_p_label, r"{\rm{Age}}/\mathrm{Gyr}"]])],
        ["rdrag", odict([
            [_p_label, r"r_\mathrm{drag}"]])]]))
# Some more, in case we want to add them at some point, described in
# https://wiki.cosmos.esa.int/planckpla2015/images/b/b9/Parameter_tag_definitions_2015.pdf
#    "zstar":       {"latex": r"z_*"},
#    "rstar":       {"latex": r"r_*"},
#    "thetastar":   {"latex": r"100\theta_*"},
#    "DAstar":      {"latex": r"D_\mathrm{A}/\mathrm{Gpc}"},
#    "zdrag":       {"latex": r"z_\mathrm{drag}"},
#    "kd":          {"latex": r"k_\mathrm{D}"},
#    "thetad":      {"latex": r"100\theta_\mathrm{D}"},
#    "zeq":         {"latex": r"z_\mathrm{eq}"},
#    "keq":         {"latex": r"k_\mathrm{eq}"},
#    "thetaeq":     {"latex": r"100\theta_\mathrm{eq}"},
#    "thetarseq":   {"latex": r"100\theta_\mathrm{s,eq}"},
for combination, info in like_cmb.items():
    if info:
        likes = ", ".join([_chi2 + _separator + like for like in info[_likelihood]])
        info[_params].update(odict([
            ["chi2__CMB", odict([[_p_derived, "lambda %s: sum([%s])" % (likes, likes)],
                                 [_p_label, r"\chi^2_\mathrm{CMB}"]])]]))

like_bao = odict([
    [_none, {}],
    ["BAO_planck_2018", {
        _desc: "Baryon acoustic oscillation data from DR12, MGS and 6DF",
        _theory: {_camb: None, _classy: None},
        _likelihood: odict([
            ["sixdf_2011_bao", None],
            ["sdss_dr7_mgs", None],
            ["sdss_dr12_consensus_bao", None]])}],
])
for combination, info in like_bao.items():
    if info:
        likes = ", ".join([_chi2 + _separator + like for like in info[_likelihood]])
        info[_params] = odict([
            ["chi2__BAO", odict([[_p_derived, "lambda %s: sum([%s])" % (likes, likes)],
                                 [_p_label, r"\chi^2_\mathrm{BAO}"]])]])

like_sn = odict([
    [_none, {}],
    ["Pantheon", {
        _desc: "Supernovae data from the Pantheon sample",
        _theory: {_camb: None, _classy: None},
        _likelihood: odict([
            ["sn_pantheon", None]])}],
])

like_H0 = odict([
    [_none, {}],
    ["Riess2018b", {
        _desc: "Local H0 measurement from Riess et al. 2018b",
        _theory: {_camb: None, _classy: None},
        _likelihood: odict([
            ["H0_riess2018b", None]])}],
])

# SAMPLERS ###############################################################################

sampler = odict([
    ["MCMC", {
        _desc: "MCMC sampler with covmat learning",
        _sampler: {"mcmc": {"covmat": "auto"}}}],
    ["PolyChord", {
        _desc: "Nested sampler, affine invariant and multi-modal",
        _sampler: {"polychord": None}}], ])

# PRESETS ################################################################################

planck_base_model = {
    "primordial": "SFSR",
    "geometry": "flat",
    "hubble": "sound_horizon_last_scattering",
    "matter": "omegab_h2, omegac_h2",
    "neutrinos": "one_heavy_planck",
    "dark_energy": "lambda",
    "bbn": "consistency",
    "reionization": "std"}
default_sampler = {"sampler": "MCMC"}

preset = odict([
    [_none, {_desc: "(No preset chosen)"}],
    # Pure CMB #######################################################
    ["planck_2015_lensing_camb", {
        _desc: "Planck 2015 + lensing with CAMB",
        "theory": _camb,
        "like_cmb": "planck_2015_lensing"}],
    ["planck_2015_lensing_classy", {
        _desc: "Planck 2015 + lensing with CLASS",
        "theory": _classy,
        "like_cmb": "planck_2015_lensing"}],
    ["planck_2015_lensing_bicep_camb", {
        _desc: "Planck 2015 + lensing + BK15 with CAMB",
        "theory": _camb,
        "primordial": "SFSR_t",
        "like_cmb": "planck_2015_lensing_bk15"}],
    ["planck_2015_lensing_bicep_classy", {
        _desc: "Planck 2015 + lensing + BK15 with CLASS",
        "theory": _classy,
        "primordial": "SFSR_t",
        "like_cmb": "planck_2015_lensing_bk15"}],
    # CMB+BAO ######################################################
    ["planck_2015_lensing_BAO_camb", {
        _desc: "Planck 2015 + lensing + BAO with CAMB",
        "theory": _camb,
        "like_cmb": "planck_2015_lensing",
        "like_bao": "BAO_planck_2018"}],
    ["planck_2015_lensing_BAO_classy", {
        _desc: "Planck 2015 + lensing + BAO with CLASS",
        "theory": _classy,
        "like_cmb": "planck_2015_lensing",
        "like_bao": "BAO_planck_2018"}],
    ["planck_2015_lensing_bicep_BAO_camb", {
        _desc: "Planck 2015 + lensing + BK15 + BAO with CAMB",
        "theory": _camb,
        "primordial": "SFSR_t",
        "like_cmb": "planck_2015_lensing_bk15",
        "like_bao": "BAO_planck_2018"}],
    ["planck_2015_lensing_bicep_BAO_classy", {
        _desc: "Planck 2015 + lensing + BK15 + BAO with CLASS",
        "theory": _classy,
        "primordial": "SFSR_t",
        "like_cmb": "planck_2015_lensing_bk15",
        "like_bao": "BAO_planck_2018"}],
    # CMB+BAO+SN ###################################################
    ["planck_2015_lensing_BAO_SN_camb", {
        _desc: "Planck 2015 + lensing + BAO + SN with CAMB",
        "theory": _camb,
        "like_cmb": "planck_2015_lensing",
        "like_bao": "BAO_planck_2018",
        "like_sn": "Pantheon"}],
    ["planck_2015_lensing_BAO_SN_classy", {
        _desc: "Planck 2015 + lensing + BAO + SN with CLASS",
        "theory": _classy,
        "like_cmb": "planck_2015_lensing",
        "like_bao": "BAO_planck_2018",
        "like_sn": "Pantheon"}],
    ["planck_2015_lensing_bicep_BAO_SN_camb", {
        _desc: "Planck 2015 + lensing + BK15 + BAO + SN with CAMB",
        "theory": _camb,
        "primordial": "SFSR_t",
        "like_cmb": "planck_2015_lensing_bk15",
        "like_bao": "BAO_planck_2018",
        "like_sn": "Pantheon"}],
    ["planck_2015_lensing_bicep_BAO_SN_classy", {
        _desc: "Planck 2015 + lensing + BK15 + BAO + SN with CLASS",
        "theory": _classy,
        "primordial": "SFSR_t",
        "like_cmb": "planck_2015_lensing_bk15",
        "like_bao": "BAO_planck_2018",
        "like_sn": "Pantheon"}],
])

# Add planck baseline model
for pre in preset.values():
    pre.update(
        {field: value for field, value in planck_base_model.items() if field not in pre})
    pre.update(default_sampler)

# BASIC INSTALLATION ######################################################################
install_basic = {
    _theory: {_camb: None, _classy: None},
    _likelihood: {
        "planck_2015_lowl": None,
        "bicep_keck_2015": None,
        "sn_pantheon": None,
        "sdss_dr12_consensus_final": None,
        "des_y1_joint": None}}



"""
.. module:: run

:Synopsis: Main scripts to run the sampling
:Author: Jesus Torrado

"""
# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division

# Global
from copy import deepcopy
from collections import OrderedDict as odict

# Local
from cobaya import __version__
from cobaya.conventions import _likelihood, _prior, _params, _theory, _sampler
from cobaya.conventions import _path_install, _debug, _debug_file, _output_prefix
from cobaya.conventions import _resume, _timing, _debug_default
from cobaya.conventions import _yaml_extensions, _separator, _full_suffix, _resume_default
from cobaya.conventions import _modules_path_arg, _force, _post
from cobaya.output import get_Output as Output
from cobaya.model import Model
from cobaya.sampler import get_sampler as Sampler
from cobaya.log import logger_setup
from cobaya.yaml import yaml_dump
from cobaya.input import get_full_info
from cobaya.mpi import import_MPI, am_single_or_primary_process
from cobaya.tools import warn_deprecation
from cobaya.post import post


def run(info):
    assert hasattr(info, "keys"), (
        "The first argument must be a dictionary with the info needed for the run. "
        "If you were trying to pass the name of an input file instead, "
        "load it first with 'cobaya.input.load_input', "
        "or, if you were passing a yaml string, load it with 'cobaya.yaml.yaml_load'.")
    # Configure the logger ASAP
    # Just a dummy import before configuring the logger, until I fix root/individual level
    import getdist
    logger_setup(info.get(_debug), info.get(_debug_file))
    import logging
    # Initialize output, if required
    output = Output(output_prefix=info.get(_output_prefix), resume=info.get(_resume),
                    force_output=info.pop(_force, None))
    # Create the full input information, including defaults for each module.
    full_info = get_full_info(info)
    if output:
        full_info[_output_prefix] = output.updated_output_prefix()
        full_info[_resume] = output.is_resuming()
    if logging.root.getEffectiveLevel() <= logging.DEBUG:
        # Don't dump unless we are doing output, just in case something not serializable
        # May be fixed in the future if we find a way to serialize external functions
        if info.get(_output_prefix) and am_single_or_primary_process():
            logging.getLogger(__name__.split(".")[-1]).info(
                "Input info updated with defaults (dumped to YAML):\n%s",
                yaml_dump(full_info))
    # TO BE DEPRECATED IN >1.2!!! #####################
    _force_reproducible = "force_reproducible"
    if _force_reproducible in info:
        info.pop(_force_reproducible)
        logging.getLogger(__name__.split(".")[-1]).warn(
            "Option '%s' is no longer necessary. Please remove it!" % _force_reproducible)
    # CHECK THAT THIS WARNING WORKS!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    ###################################################
    # We dump the info now, before modules initialization, to better track errors and
    # to check if resuming is possible asap (old and new infos are consistent)
    output.dump_info(info, full_info)
    # Initialize the posterior and the sampler
    with Model(full_info[_params], full_info[_likelihood], full_info.get(_prior),
               full_info.get(_theory), modules=info.get(_path_install),
               timing=full_info.get(_timing), allow_renames=False) as model:
        # Update the full info with the parameter routes
        keys = ([_likelihood, _theory] if _theory in full_info else [_likelihood])
        full_info.update(odict([[k,model.info()[k]] for k in keys]))
        output.dump_info(None, full_info, check_compatible=False)
        with Sampler(full_info[_sampler], model, output, resume=full_info.get(_resume),
                     modules=info.get(_path_install)) as sampler:
            sampler.run()
    # For scripted calls
    return full_info, sampler.products()


# Command-line script
def run_script():
    warn_deprecation()
    import os
    import argparse
    parser = argparse.ArgumentParser(description="Cobaya's run script.")
    parser.add_argument("input_file", nargs=1, action="store", metavar="input_file.yaml",
                        help="An input file to run.")
    parser.add_argument("-" + _modules_path_arg[0], "--" + _modules_path_arg,
                        action="store", nargs="+", metavar="/some/path", default=[None],
                        help="Path where modules were automatically installed.")
    parser.add_argument("-" + _output_prefix[0], "--" + _output_prefix,
                        action="store", nargs="+", metavar="/some/path", default=[None],
                        help="Path and prefix for the text output.")
    parser.add_argument("-" + _debug[0], "--" + _debug, action="store_true",
                        help="Produce verbose debug output.")
    continuation = parser.add_mutually_exclusive_group(required=False)
    continuation.add_argument("-" + _resume[0], "--" + _resume, action="store_true",
                              help="Resume an existing chain if it has similar info "
                                   "(fails otherwise).")
    continuation.add_argument("-" + _force[0], "--" + _force, action="store_true",
                              help="Overwrites previous output, if it exists "
                                   "(use with care!)")
    parser.add_argument("--version", action="version", version=__version__)
    args = parser.parse_args()
    if any([(os.path.splitext(f)[0] in ("input", "full")) for f in args.input_file]):
        raise ValueError("'input' and 'full' are reserved file names. "
                         "Please, use a different one.")
    load_input = import_MPI(".input", "load_input")
    given_input = args.input_file[0]
    if any(given_input.lower().endswith(ext) for ext in _yaml_extensions):
        info = load_input(given_input)
        output_prefix_cmd = getattr(args, _output_prefix)[0]
        output_prefix_input = info.get(_output_prefix)
        info[_output_prefix] = output_prefix_cmd or output_prefix_input
    else:
        # Passed an existing output_prefix? Try to find the corresponding __full.yaml
        full_file = (given_input +
                     (_separator if not given_input.endswith(os.sep) else "") +
                     _full_suffix + _yaml_extensions[0])
        try:
            info = load_input(full_file)
        except IOError:
            raise ValueError("Not a valid input file, or non-existent sample to resume")
        # We need to update the output_prefix to resume the sample *where it is*
        info[_output_prefix] = given_input
        # If input given this way, we obviously want to resume!
        info[_resume] = True
    # solve modules installation path cmd > env > input
    path_cmd = getattr(args, _modules_path_arg)[0]
    path_env = os.environ.get("COBAYA_MODULES", None)
    path_input = info.get(_path_install)
    info[_path_install] = path_cmd or (path_env or path_input)
    info[_debug] = getattr(args, _debug) or info.get(_debug, _debug_default)
    info[_resume] = getattr(args, _resume, _resume_default)
    info[_force] = getattr(args, _force, False)
    if _post in info:
        post(info)
    else:
        run(info)



"""
.. module:: cobaya.grid_tools.batchjob

:Synopsis: Classes for jobs and datasets
:Author: Antony Lewis

"""
from __future__ import absolute_import, print_function, division
import os
import shutil
import pickle
import copy
import sys
import time
import six
from getdist import types, IniFile
from getdist.mcsamples import loadMCSamples

from .conventions import _input_folder, _script_folder, _log_folder


def resetGrid(directory):
    fname = os.path.abspath(directory) + os.sep + 'batch.pyobj'
    if os.path.exists(fname): os.remove(fname)


def readobject(directory=None):
    # load this here to prevent circular
    from paramgrid import gridconfig

    if directory is None:
        directory = sys.argv[1]
    fname = os.path.abspath(directory) + os.sep + 'batch.pyobj'
    if not os.path.exists(fname):
        if gridconfig.pathIsGrid(directory):
            return gridconfig.makeGrid(directory, readOnly=True, interactive=False)
        return None
    try:
        config_dir = os.path.abspath(directory) + os.sep + 'config'
        if os.path.exists(config_dir):
            # set path in case using functions defined and hene imported from in settings file
            sys.path.insert(0, config_dir)
        with open(fname, 'rb') as inp:
            return pickle.load(inp)
    except Exception as e:
        print('Error lading cached batch object: %s', e)
        resetGrid(directory)
        if gridconfig.pathIsGrid(directory):
            return gridconfig.makeGrid(directory, readOnly=True, interactive=False)
        raise


def saveobject(obj, filename):
    with open(filename, 'wb') as output:
        pickle.dump(obj, output, pickle.HIGHEST_PROTOCOL)


def makePath(s):
    if not os.path.exists(s): os.makedirs(s)


def nonEmptyFile(fname):
    return os.path.exists(fname) and os.path.getsize(fname) > 0


def getCodeRootPath():
    return os.path.normpath(os.path.join(os.path.dirname(__file__), '..', '..')) + os.sep


class propertiesItem(object):
    def propertiesIni(self):
        if os.path.exists(self.propertiesIniFile()):
            return IniFile(self.propertiesIniFile())
        else:
            ini = IniFile()
            ini.original_filename = self.propertiesIniFile()
            return ini


class dataSet(object):
    def __init__(self, names, params=None, covmat=None, dist_settings=None):
        if not dist_settings:
            dist_settings = {}
        if isinstance(names, six.string_types): names = [names]
        if params is None:
            params = [(name + '.ini') for name in names]
        else:
            params = self.standardizeParams(params)
        if covmat is not None: self.covmat = covmat
        self.names = names
        self.params = params  # can be an array of items, either ini file name or dictionaries of parameters
        self.tag = "_".join(self.names)
        self.dist_settings = dist_settings

    def add(self, name, params=None, overrideExisting=True, dist_settings=None):
        if params is None: params = [name]
        params = self.standardizeParams(params)
        if dist_settings: self.dist_settings.update(dist_settings)
        if overrideExisting:
            self.params = params + self.params  # can be an array of items, either ini file name or dictionaries of parameters
        else:
            self.params += params
        if name is not None:
            self.names += [name]
            self.tag = "_".join(self.names)

    def addEnd(self, name, params, dist_settings=None):
        if not dist_settings:
            dist_settings = {}
        self.add(name, params, overrideExisting=False, dist_settings=dist_settings)

    def extendForImportance(self, names, params):
        data = copy.deepcopy(self)
        if not '_post_' in data.tag:
            data.tag += '_post_' + "_".join(names)
        else:
            data.tag += '_' + "_".join(names)
        data.importanceNames = names
        data.importanceParams = data.standardizeParams(params)
        data.names += data.importanceNames
        data.params += data.importanceParams
        return data

    def standardizeParams(self, params):
        if isinstance(params, dict) or isinstance(params, six.string_types): params = [params]
        for i in range(len(params)):
            if isinstance(params[i], six.string_types) and not '.ini' in params[i]: params[i] += '.ini'
        return params

    def hasName(self, name):
        if isinstance(name, six.string_types):
            return name in self.names
        else:
            return any([True for i in name if i in self.names])

    def hasAll(self, name):
        if isinstance(name, six.string_types):
            return name in self.names
        else:
            return all([(i in self.names) for i in name])

    def tagReplacing(self, x, y):
        items = []
        for name in self.names:
            if name == x:
                if y != '': items.append(y)
            else:
                items.append(name)
        return "_".join(items)

    def namesReplacing(self, dic):
        if dic is None: return self.names
        items = []
        for name in self.names:
            if name in dic:
                val = dic[name]
                if val: items.append(val)
            else:
                items.append(name)
        return items

    def makeNormedDatatag(self, dic):
        return "_".join(sorted(self.namesReplacing(dic)))


class jobGroup(object):
    def __init__(self, name, params=None, importanceRuns=None, datasets=None):
        if importanceRuns is None:
            importanceRuns = []
        if params is None:
            params = [[]]
        if datasets is None:
            datasets = []
            self.params = params
            self.groupName = name
            self.importanceRuns = importanceRuns
            self.datasets = datasets


class importanceSetting(object):
    def __init__(self, names, inis=None, dist_settings=None, minimize=True):
        if not inis:
            inis = []
        self.names = names
        self.inis = inis
        self.dist_settings = dist_settings or {}
        self.want_minimize = minimize

    def wantImportance(self, jobItem):
        return True


class importanceFilter(importanceSetting):
    # class for trivial importance sampling filters that can be done in python,
    # e.g. restricting a parameter to a new range

    def __init__(self, names, dist_settings=None, minimize=False):
        self.names = names
        self.inis = [self]
        self.dist_settings = dist_settings or {}
        self.want_minimize = minimize


class jobItem(propertiesItem):
    def __init__(self, path, param_set, data_set, base='base', minimize=True):
        self.param_set = param_set
        if not isinstance(data_set, dataSet):
            data_set = dataSet(data_set)
        self.data_set = data_set
        self.base = base
        self.paramtag = base + "_" + param_set
        self.datatag = data_set.tag
        self.name = self.paramtag + '_' + self.datatag
        self.batchPath = path
        self.relativePath = self.paramtag + os.sep + self.datatag + os.sep
        self.chainPath = path + self.relativePath
        self.chainRoot = self.chainPath + self.name
        self.distPath = self.chainPath + 'dist' + os.sep
        self.distRoot = self.distPath + self.name
        self.isImportanceJob = False
        self.importanceItems = []
        self.want_minimize = minimize
        self.result_converge = None
        self.group = None
        self.dist_settings = copy.copy(data_set.dist_settings)
        self.makeIDs()
        self.iniFile_path = _input_folder
        self.iniFile_ext = ".yaml"
        self.scriptFile_path = _script_folder
        self.logFile_path = _log_folder

    def iniFile(self, variant=''):
        if not self.isImportanceJob:
            return self.batchPath + self.iniFile_path + os.sep + self.name + variant + self.iniFile_ext
        else:
            return self.batchPath + 'postIniFiles' + os.sep + self.name + variant + self.iniFile_ext

    def propertiesIniFile(self):
        return self.chainRoot + '.properties.ini'

    def isBurnRemoved(self):
        return self.propertiesIni().bool('burn_removed')

    def makeImportance(self, importanceRuns):
        for impRun in importanceRuns:
            if isinstance(impRun, importanceSetting):
                if not impRun.wantImportance(self): continue
            else:
                if len(impRun) > 2 and not impRun[2].wantImportance(self): continue
                impRun = importanceSetting(impRun[0], impRun[1])
            if len(set(impRun.names).intersection(self.data_set.names)) > 0:
                print('importance job duplicating parent data set: %s with %s' % (self.name, impRun.names))
                continue
            data = self.data_set.extendForImportance(impRun.names, impRun.inis)
            job = jobItem(self.batchPath, self.param_set, data, minimize=impRun.want_minimize)
            job.importanceTag = "_".join(impRun.names)
            job.importanceSettings = impRun.inis
            if not '_post_' in self.name:
                tag = '_post_' + job.importanceTag
            else:
                tag = '_' + job.importanceTag
            job.name = self.name + tag
            job.chainRoot = self.chainRoot + tag
            job.distPath = self.distPath
            job.chainPath = self.chainPath
            job.relativePath = self.relativePath
            job.distRoot = self.distRoot + tag
            job.datatag = self.datatag + tag
            job.isImportanceJob = True
            job.parent = self
            job.group = self.group
            job.dist_settings.update(impRun.dist_settings)
            if isinstance(impRun, importanceFilter):
                job.importanceFilter = impRun
            job.makeIDs()
            self.importanceItems.append(job)

    def makeNormedName(self, dataSubs=None):
        normed_params = "_".join(sorted(self.param_set))
        normed_data = self.data_set.makeNormedDatatag(dataSubs)
        normed_name = self.base
        if len(normed_params) > 0: normed_name += '_' + normed_params
        normed_name += '_' + normed_data
        return normed_name, normed_params, normed_data

    def makeIDs(self):
        self.normed_name, self.normed_params, self.normed_data = self.makeNormedName()

    def matchesDatatag(self, tagList):
        if self.datatag in tagList or self.normed_data in tagList: return True
        return self.datatag.replace('_post', '') in [tag.replace('_post', '') for tag in tagList]

    def hasParam(self, name):
        if isinstance(name, six.string_types):
            return name in self.param_set
        else:
            return any([True for i in name if i in self.param_set])

    def importanceJobs(self):
        return self.importanceItems

    def importanceJobsRecursive(self):
        res = copy.copy(self.importanceItems)
        for r in self.importanceItems:
            res += r.importanceJobsRecursive()
        return res

    def removeImportance(self, job):
        if job in self.importanceItems:
            self.importanceItems.remove(job)
        else:
            for j in self.importanceItems:
                j.removeImportance(job)

    def makeChainPath(self):
        makePath(self.chainPath)
        return self.chainPath

    def writeIniLines(self, f):
        outfile = open(self.iniFile(), 'w')
        outfile.write("\n".join(f))
        outfile.close()

    def chainName(self, chain=1):
        return self.chainRoot + '_' + str(chain) + '.txt'

    def chainExists(self, chain=1):
        fname = self.chainName(chain)
        return nonEmptyFile(fname)

    def chainNames(self, num_chains=None):
        if num_chains:
            return [self.chainName(i) for i in range(num_chains)]
        else:
            i = 1
            chains = []
            while self.chainExists(i):
                chains.append(self.chainName(i))
                i += 1
            return chains

    def allChainExists(self, num_chains):
        return all([self.chainExists(i + 1) for i in range(num_chains)])

    def chainFileDate(self, chain=1):
        return os.path.getmtime(self.chainName(chain))

    def chainsDodgy(self, interval=600):
        dates = []
        i = 1
        while os.path.exists(self.chainName(i)):
            dates.append(os.path.getmtime(self.chainName(i)))
            i += 1
        return os.path.exists(self.chainName(i + 1)) or max(dates) - min(dates) > interval

    def notRunning(self):
        if not self.chainExists(): return False  # might be in queue
        lastWrite = self.chainFileDate()
        return lastWrite < time.time() - 5 * 60

    def chainMinimumExists(self):
        fname = self.chainRoot + '.minimum'
        return nonEmptyFile(fname)

    def chainBestfit(self, paramNameFile=None):
        bf_file = self.chainRoot + '.minimum'
        if nonEmptyFile(bf_file):
            return types.BestFit(bf_file, paramNameFile)
        return None

    def chainMinimumConverged(self):
        bf = self.chainBestfit()
        if bf is None: return False
        return bf.logLike < 1e29

    def convergeStat(self):
        fname = self.chainRoot + '.converge_stat'
        if not nonEmptyFile(fname): return None, None
        textFileHandle = open(fname)
        textFileLines = textFileHandle.readlines()
        textFileHandle.close()
        return float(textFileLines[0].strip()), len(textFileLines) > 1 and textFileLines[1].strip() == 'Done'

    def chainFinished(self):
        if self.isImportanceJob:
            done = self.parent.convergeStat()[1]
            if done is None or self.parentChanged() or not self.notRunning(): return False
        else:
            done = self.convergeStat()[1]
        if done is None: return False
        return done

    def wantCheckpointContinue(self, minR=0):
        R, done = self.convergeStat()
        if R is None: return False
        if not os.path.exists(self.chainRoot + '_1.chk'): return False
        return not done and R > minR

    def getDistExists(self):
        return os.path.exists(self.distRoot + '.margestats')

    def getDistNeedsUpdate(self):
        return self.chainExists() and (
                not self.getDistExists() or self.chainFileDate() > os.path.getmtime(self.distRoot + '.margestats'))

    def parentChanged(self):
        return not self.chainExists() or self.chainFileDate() < self.parent.chainFileDate()

    def R(self):
        if self.result_converge is None:
            fname = self.distRoot + '.converge'
            if not nonEmptyFile(fname): return None
            self.result_converge = types.ConvergeStats(fname)
        return float(self.result_converge.worstR())

    def hasConvergeBetterThan(self, R, returnNotExist=False):
        try:
            chainR = self.R()
            if chainR is None: return returnNotExist
            return chainR <= R
        except:
            print('WARNING: Bad .converge for ' + self.name)
            return returnNotExist

    def loadJobItemResults(self, paramNameFile=None, bestfit=True, bestfitonly=False, noconverge=False, silent=False):
        self.result_converge = None
        self.result_marge = None
        self.result_likemarge = None
        self.result_bestfit = self.chainBestfit(paramNameFile)
        if not bestfitonly:
            marge_root = self.distRoot
            if self.getDistExists():
                if not noconverge: self.result_converge = types.ConvergeStats(marge_root + '.converge')
                self.result_marge = types.MargeStats(marge_root + '.margestats', paramNameFile)
                self.result_likemarge = types.LikeStats(marge_root + '.likestats')
                if self.result_bestfit is not None and bestfit: self.result_marge.addBestFit(self.result_bestfit)
            elif not silent:
                print('missing: ' + marge_root)

    def getMCSamples(self, ini=None, settings={}):
        return loadMCSamples(self.chainRoot, jobItem=self, ini=ini, settings=settings)


class batchJob(propertiesItem):
    def __init__(self, path, cosmomcPath=None):
        self.batchPath = path
        self.skip = []
        ##        self.basePath = cosmomcPath or getCodeRootPath()
        ###        self.commonPath = os.path.join(self.basePath, iniDir)
        self.subBatches = []
        ##        self.jobItems = None
        self.getdist_options = {}
        self.iniFile_path = _input_folder
        self.scriptFile_path = _script_folder
        self.logFile_path = _log_folder

    def propertiesIniFile(self):
        return os.path.join(self.batchPath, 'config', 'config.ini')

    def makeItems(self, settings, messages=True):
        self.jobItems = []
        self.getdist_options = getattr(settings, 'getdist_options', self.getdist_options)
        allImportance = getattr(settings, 'importanceRuns', [])
        for group_name, group in settings["grid"]["groups"].items():
            for data_set in group["datasets"]:
                for param_set in group["models"]:
                    if any([data_set in (x.get("skip", {}) or {}).get(param_set, {})
                            for x in (settings["grid"], group)]):
                        continue
                    item = jobItem(self.batchPath, param_set, data_set, base=group_name)
                    if hasattr(group, 'groupName'):
                        item.group = group.groupName
                    if item.name not in self.skip:
                        item.makeImportance(group.get("importanceRuns", []))
                        item.makeImportance(allImportance)
                        self.jobItems.append(item)
        for item in getattr(settings, 'jobItems', []):
            self.jobItems.append(item)
            item.makeImportance(allImportance)
        if hasattr(settings, 'importance_filters'):
            for job in self.jobItems:
                for item in job.importanceJobs():
                    item.makeImportance(settings.importance_filters)
                job.makeImportance(settings.importance_filters)
        for item in list(self.items()):
            for x in [imp for imp in item.importanceJobsRecursive()]:
                if self.has_normed_name(x.normed_name):
                    if messages:
                        print('replacing importance sampling run '
                              'with full run: %s' % x.name)
                    item.removeImportance(x)
        for item in list(self.items()):
            for x in [imp for imp in item.importanceJobsRecursive()]:
                if self.has_normed_name(x.normed_name, wantImportance=True, exclude=x):
                    if messages:
                        print('removing duplicate importance sampling run: ' + x.name)
                    item.removeImportance(x)

    def items(self, wantSubItems=True, wantImportance=False):
        for item in self.jobItems:
            yield (item)
            if wantImportance:
                for imp in item.importanceJobsRecursive():
                    if not imp.name in self.skip:
                        yield (imp)
        if wantSubItems:
            for subBatch in self.subBatches:
                for item in subBatch.items(wantSubItems, wantImportance):
                    yield (item)

    def hasName(self, name, wantSubItems=True):
        for jobItem in self.items(wantSubItems):
            if jobItem.name == name: return True
        return False

    def has_normed_name(self, name, wantSubItems=True, wantImportance=False, exclude=None):
        return self.normed_name_item(name, wantSubItems, wantImportance, exclude) is not None

    def normed_name_item(self, name, wantSubItems=True, wantImportance=False, exclude=None):
        for jobItem in self.items(wantSubItems, wantImportance):
            if jobItem.normed_name == name and not jobItem is exclude: return jobItem
        return None

    def normalizeDataTag(self, tag):
        return "_".join(sorted(tag.replace('_post', '').split('_')))

    def resolveName(self, paramtag, datatag, wantSubItems=True, wantImportance=True, raiseError=True, base='base',
                    returnJobItem=False):
        if paramtag:
            if isinstance(paramtag, six.string_types): paramtag = paramtag.split('_')
            paramtags = [base] + sorted(paramtag)
        else:
            paramtag = [base]
            paramtags = [base]
        name = "_".join(paramtags) + '_' + self.normalizeDataTag(datatag)
        jobItem = self.normed_name_item(name, wantSubItems, wantImportance)
        if jobItem is not None: return (jobItem.name, jobItem)[returnJobItem]
        if raiseError:
            raise Exception('No match for paramtag, datatag... ' + "_".join(paramtag) + ', ' + datatag)
        else:
            return None

    def resolveRoot(self, root):
        for jobItem in self.items(True, True):
            if jobItem.name == root: return jobItem
        return self.normed_name_item(root, True, True)

    def save(self, filename=''):
        saveobject(self, (self.batchPath + 'batch.pyobj', filename)[filename != ''])

    def makeDirectories(self, setting_file=None):
        makePath(self.batchPath)
        if setting_file:
            makePath(self.batchPath + 'config')
            setting_file = setting_file.replace('.pyc', '.py')
            shutil.copy(setting_file, self.batchPath + 'config')
            props = self.propertiesIni()
            props.params['setting_file'] = os.path.split(setting_file)[-1]
            props.saveFile()
        makePath(self.batchPath + self.iniFile_path)
        makePath(self.batchPath + self.scriptFile_path)
        makePath(self.batchPath + self.logFile_path)


####        makePath(self.batchPath + 'postIniFiles')


class batchJob_cosmomc(propertiesItem):
    def __init__(self, path, iniDir, cosmomcPath=None):
        self.batchPath = path
        self.skip = []
        self.basePath = cosmomcPath or getCodeRootPath()
        self.commonPath = os.path.join(self.basePath, iniDir)
        self.subBatches = []
        self.jobItems = None
        self.getdist_options = {}

    def propertiesIniFile(self):
        return os.path.join(self.batchPath, 'config', 'config.ini')

    def makeItems(self, settings, messages=True):
        self.jobItems = []
        self.getdist_options = getattr(settings, 'getdist_options', self.getdist_options)
        allImportance = getattr(settings, 'importanceRuns', [])
        for group in settings.groups:
            for data_set in group.datasets:
                for param_set in group.params:
                    item = jobItem(self.batchPath, param_set, data_set)
                    if hasattr(group, 'groupName'): item.group = group.groupName
                    if not item.name in self.skip:
                        item.makeImportance(group.importanceRuns)
                        item.makeImportance(allImportance)
                        self.jobItems.append(item)
        for item in getattr(settings, 'jobItems', []):
            self.jobItems.append(item)
            item.makeImportance(allImportance)

        if hasattr(settings, 'importance_filters'):
            for job in self.jobItems:
                for item in job.importanceJobs():
                    item.makeImportance(settings.importance_filters)
                job.makeImportance(settings.importance_filters)

        for item in list(self.items()):
            for x in [imp for imp in item.importanceJobsRecursive()]:
                if self.has_normed_name(x.normed_name):
                    if messages: print('replacing importance sampling run with full run: ' + x.name)
                    item.removeImportance(x)
        for item in list(self.items()):
            for x in [imp for imp in item.importanceJobsRecursive()]:
                if self.has_normed_name(x.normed_name, wantImportance=True, exclude=x):
                    if messages: print('removing duplicate importance sampling run: ' + x.name)
                    item.removeImportance(x)

    def items(self, wantSubItems=True, wantImportance=False):
        for item in self.jobItems:
            yield (item)
            if wantImportance:
                for imp in item.importanceJobsRecursive():
                    if not imp.name in self.skip: yield (imp)

        if wantSubItems:
            for subBatch in self.subBatches:
                for item in subBatch.items(wantSubItems, wantImportance): yield (item)

    def hasName(self, name, wantSubItems=True):
        for jobItem in self.items(wantSubItems):
            if jobItem.name == name: return True
        return False

    def has_normed_name(self, name, wantSubItems=True, wantImportance=False, exclude=None):
        return self.normed_name_item(name, wantSubItems, wantImportance, exclude) is not None

    def normed_name_item(self, name, wantSubItems=True, wantImportance=False, exclude=None):
        for jobItem in self.items(wantSubItems, wantImportance):
            if jobItem.normed_name == name and not jobItem is exclude: return jobItem
        return None

    def normalizeDataTag(self, tag):
        return "_".join(sorted(tag.replace('_post', '').split('_')))

    def resolveName(self, paramtag, datatag, wantSubItems=True, wantImportance=True, raiseError=True, base='base',
                    returnJobItem=False):
        if paramtag:
            if isinstance(paramtag, six.string_types): paramtag = paramtag.split('_')
            paramtags = [base] + sorted(paramtag)
        else:
            paramtag = [base]
            paramtags = [base]
        name = "_".join(paramtags) + '_' + self.normalizeDataTag(datatag)
        jobItem = self.normed_name_item(name, wantSubItems, wantImportance)
        if jobItem is not None: return (jobItem.name, jobItem)[returnJobItem]
        if raiseError:
            raise Exception('No match for paramtag, datatag... ' + "_".join(paramtag) + ', ' + datatag)
        else:
            return None

    def resolveRoot(self, root):
        for jobItem in self.items(True, True):
            if jobItem.name == root: return jobItem
        return self.normed_name_item(root, True, True)

    def save(self, filename=''):
        saveobject(self, (self.batchPath + 'batch.pyobj', filename)[filename != ''])

    def makeDirectories(self, setting_file=None):
        makePath(self.batchPath)
        if setting_file:
            makePath(self.batchPath + 'config')
            setting_file = setting_file.replace('.pyc', '.py')
            shutil.copy(setting_file, self.batchPath + 'config')
            props = self.propertiesIni()
            props.params['setting_file'] = os.path.split(setting_file)[-1]
            props.saveFile()
        makePath(self.batchPath + 'iniFiles')
        makePath(self.batchPath + 'postIniFiles')



"""
.. module:: parameterization

:Synopsis: Class managing the possibly different parameterizations
           used by sampler and likelihoods
:Author: Jesus Torrado

"""
# Python 2/3 compatibility
from __future__ import absolute_import
from __future__ import division
from six import string_types

# Global
from collections import OrderedDict as odict
from numbers import Number
from itertools import chain
from copy import deepcopy

# Local
from cobaya.conventions import _prior, _p_drop, _p_derived, _p_label, _p_value, _p_renames
from cobaya.tools import get_external_function, ensure_nolatex, is_valid_variable_name, getargspec
from cobaya.log import HandledException

# Logger
import logging

log = logging.getLogger(__name__.split(".")[-1])


def is_fixed_param(info_param):
    """
    Returns True if the parameter has been fixed to a value or through a function.
    """
    return expand_info_param(info_param).get(_p_value, None) is not None


def is_sampled_param(info_param):
    """
    Returns True if the parameter has a prior.
    """
    return _prior in expand_info_param(info_param)


def is_derived_param(info_param):
    """
    Returns True if the parameter is saved as a derived one.
    """
    return expand_info_param(info_param).get(_p_derived, False)


def expand_info_param(info_param):
    """
    Expands the info of a parameter, from the user friendly, shorter format
    to a more unambiguous one.
    """
    info_param = deepcopy(info_param)
    if not hasattr(info_param, "keys"):
        if info_param is None:
            info_param = odict()
        else:
            info_param = odict([[_p_value, info_param]])
    if all([(f not in info_param) for f in [_prior, _p_value, _p_derived]]):
        info_param[_p_derived] = True
    # Dynamical input parameters: save as derived by default
    value = info_param.get(_p_value, None)
    if isinstance(value, string_types) or callable(value):
        info_param[_p_derived] = info_param.get(_p_derived, True)
    return info_param


def reduce_info_param(info_param):
    """
    Compresses the info of a parameter, suppressing default values.
    This is the opposite of :func:`~input.expand_info_param`.
    """
    info_param = deepcopy(info_param)
    if not hasattr(info_param, "keys"):
        return
    # All parameters without a prior are derived parameters unless otherwise specified
    if info_param.get(_p_derived) is True:
        info_param.pop(_p_derived)
    # Fixed parameters with single "value" key
    if list(info_param) == [_p_value]:
        return info_param[_p_value]
    return info_param


def call_param_func(p, func, kwargs):
    try:
        return func(**kwargs)
    except NameError as exception:
        unknown = str(exception).split("'")[1]
        log.error("Unknown variable '%s' was referenced in the definition of "
                  "the parameter '%s', with arguments %r.", unknown, p, list(kwargs))
        raise HandledException
    except:
        log.error("Function for parameter '%s' failed at evaluation "
                  "and threw the following exception:", p)
        raise


class Parameterization(object):
    """
    Class managing parameterization.
    Translates parameter between sampler+prior and likelihood
    """

    def __init__(self, info_params, allow_renames=True, ignore_unused_sampled=False):
        self.allow_renames = allow_renames
        # First, we load the parameters,
        # not caring about whether they are understood by any likelihood.
        # `input` contains the parameters (expected to be) understood by the likelihood,
        #   with its fixed value, its fixing function, or None if their value is given
        #   directly by the sampler.
        self._input = odict()
        self._input_funcs = dict()
        self._input_args = dict()
        self._output = odict()
        self._constant = odict()
        self._sampled = odict()
        self._sampled_info = odict()
        self._sampled_renames = odict()
        self._derived = odict()
        self._derived_funcs = dict()
        self._derived_args = dict()
        # Notice here that expand_info_param *always* adds a _p_derived:True tag
        # to infos without _prior or _p_value, and a _p_value field to fixed params
        for p, info in info_params.items():
            if is_fixed_param(info):
                if isinstance(info[_p_value], Number):
                    self._constant[p] = info[_p_value]
                    if not info.get(_p_drop, False):
                        self._input[p] = self._constant[p]
                else:
                    self._input[p] = None
                    self._input_funcs[p] = get_external_function(info[_p_value])
                    self._input_args[p] = getargspec(self._input_funcs[p]).args
            if is_sampled_param(info):
                self._sampled[p] = None
                self._sampled_info[p] = deepcopy(info)
                if not info.get(_p_drop, False):
                    self._input[p] = None
                self._sampled_renames[p] = (
                    (lambda x: [x] if isinstance(x, string_types) else x)
                    (info.get(_p_renames, [])))
            if is_derived_param(info):
                self._derived[p] = deepcopy(info)
                # Dynamical parameters whose value we want to save
                if info[_p_derived] is True and is_fixed_param(info):
                    info[_p_derived] = "lambda %s: %s" % (p, p)
                if info[_p_derived] is True:
                    self._output[p] = None
                else:
                    self._derived_funcs[p] = get_external_function(info[_p_derived])
                    self._derived_args[p] = getargspec(self._derived_funcs[p]).args
        # Check that the sampled and derived params are all valid python variable names
        for p in chain(self.sampled_params(), self.derived_params()):
            if not is_valid_variable_name(p):
                is_in = p in self.sampled_params()
                eg_in = "  p_prime:\n    prior: ...\n  %s: 'lambda p_prime: p_prime'\n" % p
                eg_out = "  p_prime: 'lambda %s: %s'\n" % (p, p)
                log.error("Parameter name '%s' is not a valid Python variable name "
                          "(it needs to start with a letter or '_').\n"
                          "If this is an %s parameter of a likelihood or theory, "
                          "whose name you cannot change,%s define an associated "
                          "%s one with a valid name 'p_prime' as: \n\n%s",
                          p, "input" if is_in else "output",
                          "" if is_in else " remove it and",
                          "sampled" if is_in else "derived",
                          eg_in if is_in else eg_out)
                raise HandledException
        # Assume that the *un*known function arguments are likelihood output parameters
        args = (set(chain(*self._input_args.values()))
                .union(chain(*self._derived_args.values())))
        for p in (list(self._constant) + list(self._input) +
                  list(self._sampled) + list(self._derived)):
            if p in args:
                args.remove(p)
        self._output.update({p: None for p in args})
        # Useful sets: directly-sampled input parameters and directly "output-ed" derived
        self._directly_sampled = [p for p in self._input if p in self._sampled]
        self._directly_output = [p for p in self._derived if p in self._output]
        # Useful mapping: input params that vary if each sampled is varied
        self._sampled_input_dependence = odict(
            [[s, [i for i in self._input if s in self._input_args.get(i, {})]]
             for s in self._sampled])
        # From here on, some error control.
        dropped_but_never_used = (
            set([p for p, v in self._sampled_input_dependence.items() if not v])
                .difference(set(self._directly_sampled)))
        if dropped_but_never_used and not ignore_unused_sampled:
            log.error("Parameters %r are sampled but not passed to the likelihood or "
                      "theory code, neither ever used as arguments for any parameters. "
                      "Check that you are not using the '%s' tag unintentionally.",
                      list(dropped_but_never_used), _p_drop)
            raise HandledException
        # input params depend on input and sampled only, never on output/derived
        all_input_arguments = set(chain(*self._input_args.values()))
        bad_input_dependencies = all_input_arguments.difference(
            set(self.input_params()).union(set(self.sampled_params())).union(set(self.constant_params())))
        if bad_input_dependencies:
            log.error("Input parameters defined as functions can only depend on other "
                      "input parameters that are not defined as functions. "
                      "In particular, an input parameter cannot depend on %r",
                      list(bad_input_dependencies))
            raise HandledException

    def input_params(self):
        return deepcopy(self._input)

    def output_params(self):
        return deepcopy(self._output)

    def constant_params(self):
        return deepcopy(self._constant)

    def sampled_params(self):
        return deepcopy(self._sampled)

    def sampled_params_info(self):
        return deepcopy(self._sampled_info)

    def sampled_params_renames(self):
        return deepcopy(self._sampled_renames)

    def derived_params(self):
        return deepcopy(self._derived)

    def sampled_input_dependence(self):
        return deepcopy(self._sampled_input_dependence)

    def _to_input(self, sampled_params_values):
        # Store sampled params, so that derived can depend on them
        if not hasattr(sampled_params_values, "keys"):
            sampled_params_values = odict(
                zip(self.sampled_params(), sampled_params_values))
        elif not isinstance(sampled_params_values, odict):
            sampled_params_values = odict(
                [(p, sampled_params_values[p]) for p in self.sampled_params()])
        self._sampled = deepcopy(sampled_params_values)
        # Fill first directly sampled input parameters
        self._input.update(
            {p: sampled_params_values[p] for p in self._directly_sampled})
        # Then evaluate the functions
        resolved_old = None
        resolved = []
        while resolved != resolved_old:
            resolved_old = deepcopy(resolved)
            for p in self._input_funcs:
                if p in resolved:
                    continue
                args = {p:
                    self._constant.get(
                        p, self._input.get(
                            p, sampled_params_values.get(p, None)))
                    for p in self._input_args[p]}
                if not all([isinstance(v, Number) for v in args.values()]):
                    continue
                self._input[p] = call_param_func(p, self._input_funcs[p], args)
                resolved.append(p)
        if set(resolved) != set(self._input_funcs):
            log.error("Could not resolve arguments for input parameters %s. Maybe there "
                      "is a circular dependency between derived parameters?",
                      list(set(self._input_funcs).difference(set(resolved))))
            raise HandledException
        return self.input_params()

    def _to_derived(self, output_params_values):
        if not hasattr(output_params_values, "keys"):
            output_params_values = dict(
                zip(self.output_params(), output_params_values))
        # Fill first derived parameters which are direct output parameters
        self._derived.update(
            {p: output_params_values[p] for p in self._directly_output})
        # Then evaluate the functions
        resolved_old = None
        resolved = []
        while resolved != resolved_old:
            resolved_old = deepcopy(resolved)
            for p in self._derived_funcs:
                if p in resolved:
                    continue
                args = {p: (self.input_params().get(
                    p, self.sampled_params().get(p, output_params_values.get(
                        p, self._derived.get(p, None)))))
                    for p in self._derived_args[p]}
                if not all([isinstance(v, Number) for v in args.values()]):
                    continue
                self._derived[p] = call_param_func(p, self._derived_funcs[p], args)
                resolved.append(p)
        if set(resolved) != set(self._derived_funcs):
            log.error("Could not resolve arguments for derived parameters %s. Maybe there"
                      " is a circular dependency between derived parameters?",
                      list(set(self._derived_funcs).difference(set(resolved))))
            raise HandledException
        return list(self._derived.values())

    def _check_sampled(self, **sampled_params):
        """
        Check that the input dictionary contains all the sampled parameters,
        and just them. Is aware of known renamings.

        Returns `OrderedDict` of parameters (model's naming) and their values.
        """
        sampled_output = odict()
        sampled_input = deepcopy(sampled_params)
        for p, renames in self._sampled_renames.items():
            for pprime in sampled_input:
                if pprime == p or (pprime in renames if self.allow_renames else False):
                    sampled_output[p] = sampled_input.pop(pprime)
                    break
        if len(sampled_output) < len(self._sampled):
            not_found = set(self.sampled_params()).difference(set(sampled_output))
            if self.allow_renames:
                log.error("The following expected sampled parameters " +
                          ("(or their aliases) " if self.allow_renames else "") +
                          "where not found : %r",
                          ({p: self._sampled_renames[p] for p in not_found}
                           if self.allow_renames else not_found))
            else:
                log.error("The following expected sampled parameters "
                          "where not found : %r",
                          {p: self._sampled_renames[p] for p in not_found})
            raise HandledException
        if sampled_input:
            not_used = set(sampled_input)
            duplicated = not_used.intersection(set(
                chain(*[list(chain(*[[k], v])) for k, v in self._sampled_renames.items()])))
            not_used = not_used.difference(duplicated)
            derived = not_used.intersection(set(self.derived_params()))
            input_ = not_used.intersection(set(self.input_params()))
            unknown = not_used.difference(derived).difference(input_)
            log.error(
                "Incorrect parameters! " +
                ("\n   Duplicated entries (using their aliases): %r" % list(duplicated)
                 if duplicated else "") +
                ("\n   Not known: %r" % list(unknown) if unknown else "") +
                ("\n   Cannot be fixed: %r " % list(input_) +
                 "--> instead, fix sampled parameters that depend on them!"
                 if input_ else "") +
                ("\n   Cannot be fixed because are derived parameters: %r " % list(derived)
                 if derived else ""))
            raise HandledException
        return sampled_output

    def labels(self):
        """
        Returns a dictionary of LaTeX labels of the sampled and derived parameters.

        Uses the parameter name of no label has been given.
        """
        get_label = lambda p, info: (
            ensure_nolatex(getattr(info, "get", lambda x, y: y)(_p_label, p)))
        return odict([[p, get_label(p, info)] for p, info in
                      list(self.sampled_params().items()) +
                      list(self.derived_params().items())])

    # Python magic for the "with" statement
    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        return



"""
.. module:: _sn_prototype

:Synopsis: Supernovae likelihood, from CosmoMC's JLA module, for Pantheon and JLA samples.
:Author: Alex Conley, Marc Betoule, Antony Lewis (see source for more specific authorship)

This code provides the following likelihoods:

- ``sn_pantheon``, for the Pantheon SN Ia sample (including Pan-STARRS1 MDS and others)
- ``sn_jla``, for the JLA SN Ia sample, based on joint SNLS/SDSS SN Ia data
- ``sn_jla_lite``, an alternative version of ``sn_jla``, marginalized over
  nuisance parameters

.. |br| raw:: html

   <br />

.. note::

   - If you use ``sn_pantheon``, please cite:|br|
     Scolnic, D. M. et al,
     `The Complete Light-curve Sample of Spectroscopically
     Confirmed Type Ia Supernovae from Pan-STARRS1 and
     Cosmological Constraints from The Combined Pantheon Sample`
     `(arXiv:1710.00845) <https://arxiv.org/abs/1710.00845>`_
   - If you use ``sn_jla`` or ``sn_jla_lite``, please cite:|br|
     Betoule, M. et al,
     `Improved cosmological constraints from a joint analysis
     of the SDSS-II and SNLS supernova samples`
     `(arXiv:1401.4064) <https://arxiv.org/abs/1401.4064>`_


Usage
-----

To use any of these likelihoods, simply mention them in the likelihoods block
(do not use ``sn_jla`` and its `lite` version simultaneously), or add them
using the :doc:`input generator <cosmo_basic_runs>`.

The settings for each likelihood, as well as the nuisance parameters and their default
priors (in the ``sn_jla`` case only) can be found in the ``defaults.yaml``
files in the folder for the source code of each of these likelihoods,
and are reproduced below.

You shouldn't need to modify any of the options of these likelihoods,
but if you really need to, just copy the ``likelihood`` block into your input ``yaml``
file and modify whatever options you want (you can delete the rest).

.. literalinclude:: ../cobaya/likelihoods/sn_pantheon/sn_pantheon.yaml
   :language: yaml

.. literalinclude:: ../cobaya/likelihoods/sn_jla/sn_jla.yaml
   :language: yaml

.. literalinclude:: ../cobaya/likelihoods/sn_jla_lite/sn_jla_lite.yaml
   :language: yaml


Installation
------------

This likelihood can be installed automatically as explained in :doc:`installation_cosmo`.
If are following the instructions there (you should!), you don't need to read the rest
of this section.

Manual installation of the SN Ia likelihoods data
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Assuming you are installing all your
likelihoods under ``/path/to/likelihoods``, simply do

.. code:: bash

   $ cd /path/to/likelihoods
   $ git clone https://github.com/JesusTorrado/sn_data.git

After this, mention the path to this likelihood when you include it in an input file as

.. code-block:: yaml

   likelihood:
     sn_[pantheon|jla|jla_lite]:
       path: /path/to/likelihoods/sn_data

"""

# Supernovae likelihood, from CosmoMC's JLA module. For Pantheon and JLA Supernovae,
#  History:
#  Written by Alex Conley, Dec 2006
#   aconley, Jan 2007: The OpenMP stuff was causing massive slowdowns on
#      some processors (ones with hyperthreading), so it was removed
#   aconley, Jul 2009: Added absolute distance support
#   aconley, May 2010: Added twoscriptm support
#   aconley, Apr 2011: Fix some non standard F90 usage.  Thanks to
#                       Zhiqi Huang for catching this.
#   aconley, April 2011: zhel, zcmb read in wrong order.  Thanks to
#                       Xiao Dong-Li and Shuang Wang for catching this
#   mbetoule, Dec 2013: adaptation to the JLA sample
#   AL, Mar 2014: updates for latest CosmoMC structure
#   AL, June 2014: updated JLA_marginalize=T handling so it should work
#   AL, March 2018: this python version

# Global
from __future__ import division, print_function
import numpy as np
import io
import os
import logging
from getdist import IniFile

# Local
from cobaya.conventions import _package, _path_install
from cobaya.likelihood import Likelihood
from cobaya.log import HandledException
from cobaya.install import download_github_release

_twopi = 2 * np.pi


class _sn_prototype(Likelihood):

    def initialize(self):
        def relative_path(tag):
            return ini.relativeFileName(tag).replace('data/', '').replace('Pantheon/', '')

        # has_absdist = F, intrinsicdisp=0, idispdataset=False
        if not self.path:
            if self.path_install:
                from importlib import import_module
                self.path = getattr(
                    import_module(
                        _package + ".likelihoods." + self.name, package=_package),
                    "get_path")(self.path_install)
            else:
                self.log.error("No path given to the %s likelihood. Set the likelihood"
                               " property 'path' or the common property '%s'.",
                               self.dataset_file, _path_install)
                raise HandledException
        self.path = os.path.normpath(self.path)
        self.dataset_file_path = os.path.normpath(os.path.join(self.path, self.dataset_file))
        self.log.info("Reading data from %s", self.dataset_file_path)
        if not os.path.exists(self.dataset_file_path):
            self.log.error("The likelihood is not installed in the given path: "
                           "cannot find the file '%s'.", self.dataset_file_path)
            raise HandledException
        ini = IniFile(self.dataset_file_path)
        ini.params.update(self.dataset_params or {})
        self.twoscriptmfit = ini.bool('twoscriptmfit')
        if self.twoscriptmfit:
            scriptmcut = ini.float('scriptmcut', 10.)
        assert not ini.float('intrinsicdisp', 0) and not ini.float('intrinsicdisp0', 0)
        if hasattr(self, "alpha_beta_names"):
            self.alpha_name = self.alpha_beta_names[0]
            self.beta_name = self.alpha_beta_names[1]
        self.pecz = ini.float('pecz', 0.001)
        cols = None
        self.has_third_var = False
        data_file = os.path.join(self.path, ini.string("data_file"))
        self.log.debug('Reading %s' % data_file)
        supernovae = {}
        self.names = []
        ix = 0
        with io.open(data_file, 'r') as f:
            lines = f.readlines()
            for line in lines:
                if '#' in line:
                    cols = line[1:].split()
                    for rename, new in zip(
                            ['mb', 'color', 'x1', '3rdvar', 'd3rdvar',
                             'cov_m_s', 'cov_m_c', 'cov_s_c'],
                            ['mag', 'colour', 'stretch', 'third_var',
                             'dthird_var', 'cov_mag_stretch',
                             'cov_mag_colour', 'cov_stretch_colour']):
                        if rename in cols:
                            cols[cols.index(rename)] = new
                    self.has_third_var = 'third_var' in cols
                    zeros = np.zeros(len(lines) - 1)
                    self.third_var = zeros.copy()
                    self.dthird_var = zeros.copy()
                    self.set = zeros.copy()
                    for col in cols:
                        setattr(self, col, zeros.copy())
                elif line.strip():
                    if cols is None:
                        self.log.error('Data file must have comment header')
                        raise HandledException
                    vals = line.split()
                    for i, (col, val) in enumerate(zip(cols, vals)):
                        if col == 'name':
                            supernovae[val] = ix
                            self.names.append(val)
                        else:
                            getattr(self, col)[ix] = np.float64(val)
                    ix += 1
        self.z_var = self.dz ** 2
        self.mag_var = self.dmb ** 2
        self.stretch_var = self.dx1 ** 2
        self.colour_var = self.dcolor ** 2
        self.thirdvar_var = self.dthird_var ** 2
        self.nsn = ix
        self.log.debug('Number of SN read: %s ' % self.nsn)
        if self.twoscriptmfit and not self.has_third_var:
            self.log.error('twoscriptmfit was set but thirdvar information not present')
            raise HandledException
        if ini.bool('absdist_file'):
            self.log.error('absdist_file not supported')
            raise HandledException
        covmats = [
            'mag', 'stretch', 'colour', 'mag_stretch', 'mag_colour', 'stretch_colour']
        self.covs = {}
        for name in covmats:
            if ini.bool('has_%s_covmat' % name):
                self.log.debug('Reading covmat for: %s ' % name)
                self.covs[name] = self._read_covmat(
                    os.path.join(self.path, ini.string('%s_covmat_file' % name)))
        self.alphabeta_covmat = (len(self.covs.items()) > 1 or
                                 self.covs.get('mag', None) is None)
        self._last_alpha = np.inf
        self._last_beta = np.inf
        self.marginalize = getattr(self, "marginalize", False)
        assert self.covs
        # jla_prep
        zfacsq = 25.0 / np.log(10.0) ** 2
        self.pre_vars = self.mag_var + zfacsq * self.pecz ** 2 * (
                (1.0 + self.zcmb) / (self.zcmb * (1 + 0.5 * self.zcmb))) ** 2
        if self.twoscriptmfit:
            A1 = np.zeros(self.nsn)
            A2 = np.zeros(self.nsn)
            A1[self.third_var <= scriptmcut] = 1
            A2[self.third_var > scriptmcut] = 1
            has_A1 = np.any(A1)
            has_A2 = np.any(A2)
            if not has_A1:
                # swap
                A1 = A2
                A2 = np.zeros(self.nsn)
                has_A2 = False
            if not has_A2:
                self.twoscriptmfit = False
            self.A1 = A1
            self.A2 = A2
        if self.marginalize:
            self.step_width_alpha = self.marginalize_params['step_width_alpha']
            self.step_width_beta = self.marginalize_params['step_width_beta']
            _marge_steps = self.marginalize_params['marge_steps']
            self.alpha_grid = np.empty((2 * _marge_steps + 1) ** 2)
            self.beta_grid = self.alpha_grid.copy()
            _int_points = 0
            for alpha_i in range(-_marge_steps, _marge_steps + 1):
                for beta_i in range(-_marge_steps, _marge_steps + 1):
                    if alpha_i ** 2 + beta_i ** 2 <= _marge_steps ** 2:
                        self.alpha_grid[_int_points] = (
                                self.marginalize_params['alpha_centre'] +
                                alpha_i * self.step_width_alpha)
                        self.beta_grid[_int_points] = (
                                self.marginalize_params['beta_centre'] +
                                beta_i * self.step_width_beta)
                        _int_points += 1
            self.log.debug('Marignalizing alpha, beta over %s points' % _int_points)
            self.marge_grid = np.empty(_int_points)
            self.int_points = _int_points
            self.alpha_grid = self.alpha_grid[:_int_points]
            self.beta_grid = self.beta_grid[:_int_points]
            self.invcovs = np.empty(_int_points, dtype=np.object)
            if self.precompute_covmats:
                for i, (alpha, beta) in enumerate(zip(self.alpha_grid, self.beta_grid)):
                    self.invcovs[i] = self.inverse_covariance_matrix(alpha, beta)
        elif not self.alphabeta_covmat:
            self.inverse_covariance_matrix()

    def add_theory(self):
        # State requisites to the theory code
        self.theory.needs(**{"angular_diameter_distance": {"z": self.zcmb}})

    def _read_covmat(self, filename):
        cov = np.loadtxt(filename)
        if np.isscalar(cov[0]) and cov[0] ** 2 + 1 == len(cov):
            cov = cov[1:]
        return cov.reshape((self.nsn, self.nsn))

    def inverse_covariance_matrix(self, alpha=0, beta=0):
        if 'mag' in self.covs:
            invcovmat = self.covs['mag'].copy()
        else:
            invcovmat = 0
        if self.alphabeta_covmat:
            if np.isclose(alpha, self._last_alpha) and np.isclose(beta, self._last_beta):
                return self.invcov
            self._last_alpha = alpha
            self._last_beta = beta
            alphasq = alpha * alpha
            betasq = beta * beta
            alphabeta = alpha * beta
            if 'stretch' in self.covs:
                invcovmat += alphasq * self.covs['stretch']
            if 'colour' in self.covs:
                invcovmat += betasq * self.covs['colour']
            if 'mag_stretch' in self.covs:
                invcovmat += 2 * alpha * self.covs['mag_stretch']
            if 'mag_colour' in self.covs:
                invcovmat -= 2 * beta * self.covs['mag_colour']
            if 'stretch_colour' in self.covs:
                invcovmat -= 2 * alphabeta * self.covs['stretch_colour']
            delta = (self.pre_vars + alphasq * self.stretch_var +
                     betasq * self.colour_var + 2.0 * alpha * self.cov_mag_stretch +
                     -2.0 * beta * self.cov_mag_colour +
                     -2.0 * alphabeta * self.cov_stretch_colour)
        else:
            delta = self.pre_vars
        np.fill_diagonal(invcovmat, invcovmat.diagonal() + delta)
        self.invcov = np.linalg.inv(invcovmat)
        return self.invcov

    def alpha_beta_logp(self, lumdists, alpha=0, beta=0, invcovmat=None):
        if self.alphabeta_covmat:
            alphasq = alpha * alpha
            betasq = beta * beta
            alphabeta = alpha * beta
            invvars = 1.0 / (self.pre_vars + alphasq * self.stretch_var +
                             betasq * self.colour_var +
                             2.0 * alpha * self.cov_mag_stretch -
                             2.0 * beta * self.cov_mag_colour -
                             2.0 * alphabeta * self.cov_stretch_colour)
            wtval = np.sum(invvars)
            estimated_scriptm = np.sum((self.mag - lumdists) * invvars) / wtval
            diffmag = (self.mag - lumdists + alpha * self.stretch -
                       beta * self.colour - estimated_scriptm)
            if invcovmat is None:
                invcovmat = self.inverse_covariance_matrix(alpha, beta)
        else:
            invvars = 1.0 / self.pre_vars
            wtval = np.sum(invvars)
            estimated_scriptm = np.sum((self.mag - lumdists) * invvars) / wtval
            diffmag = self.mag - lumdists - estimated_scriptm
            invcovmat = self.invcov
        invvars = invcovmat.dot(diffmag)
        amarg_A = invvars.dot(diffmag)
        if self.twoscriptmfit:
            # could simplify this..
            amarg_B = invvars.dot(self.A1)
            amarg_C = invvars.dot(self.A2)
            invvars = invcovmat.dot(self.A1)
            amarg_D = invvars.dot(self.A2)
            amarg_E = invvars.dot(self.A1)
            invvars = invcovmat.dot(self.A2)
            amarg_F = invvars.dot(self.A2)
            tempG = amarg_F - amarg_D * amarg_D / amarg_E
            assert tempG >= 0
            chi2 = (amarg_A + np.log(amarg_E / _twopi) +
                    np.log(tempG / _twopi) - amarg_C * amarg_C / tempG -
                    amarg_B * amarg_B * amarg_F / (amarg_E * tempG) +
                    2.0 * amarg_B * amarg_C * amarg_D / (amarg_E * tempG))
        else:
            amarg_B = np.sum(invvars)
            amarg_E = np.sum(invcovmat)
            chi2 = amarg_A + np.log(amarg_E / _twopi) - amarg_B ** 2 / amarg_E
        return - chi2 / 2

    def logp(self, **params_values):
        angular_diameter_distances = self.theory.get_angular_diameter_distance(self.zcmb)
        lumdists = (5 * np.log10((1 + self.zhel) * (1 + self.zcmb) *
                                 angular_diameter_distances))
        if self.marginalize:
            # Should parallelize this loop
            for i in range(self.int_points):
                self.marge_grid[i] = - self.alpha_beta_logp(
                    lumdists, self.alpha_grid[i],
                    self.beta_grid[i], invcovmat=self.invcovs[i])
            grid_best = np.min(self.marge_grid)
            return - grid_best + np.log(
                np.sum(np.exp(- self.marge_grid[self.marge_grid != np.inf] + grid_best)) *
                self.step_width_alpha * self.step_width_beta)
        else:
            if self.alphabeta_covmat:
                return self.alpha_beta_logp(lumdists, params_values[self.alpha_name],
                                            params_values[self.beta_name])
            else:
                return self.alpha_beta_logp(lumdists)


# Installation routines ##################################################################

# name of the data and covmats repo/folder
sn_data_name = "sn_data"
sn_data_version = "v1.3"


def get_path(path):
    return os.path.realpath(os.path.join(path, "data", sn_data_name))


def is_installed(**kwargs):
    return os.path.exists(os.path.realpath(
        os.path.join(kwargs["path"], "data", sn_data_name)))


def install(path=None, force=False, code=False, data=True, no_progress_bars=False):
    if not data:
        return True
    log = logging.getLogger(__name__.split(".")[-1])
    log.info("Downloading Supernovae data...")
    return download_github_release(os.path.join(path, "data"), sn_data_name,
                                   sn_data_version, no_progress_bars=no_progress_bars)



"""
.. module:: sampler

:Synopsis: Prototype sampler class and sampler loader
:Author: Jesus Torrado

cobaya includes by default a
:doc:`Monte Carlo Markov Chain (MCMC) sampler <sampler_mcmc>`
(a direct translation from `CosmoMC <https://cosmologist.info/cosmomc/>`_) and a dummy
:doc:`evaluate <sampler_evaluate>` sampler that simply evaluates the posterior at a given
(or sampled) reference point. It also includes an interface to the
:doc:`PolyChord sampler <sampler_polychord>` (needs to be installed separately).

The sampler to use is specified by a `sampler` block in the input file, whose only member
is the sampler used, containing some options, if necessary.

.. code-block:: yaml

   sampler:
     mcmc:
       max_samples: 1000

or

.. code-block:: yaml

   sampler:
     polychord:
       path: /path/to/cosmo/PolyChord

Samplers can in general be swapped in the input file without needing to modify any other
block of the input.

In the cobaya code tree, each sampler is placed in its own folder, containing a file
defining the sampler's class, which inherits from the :class:`cobaya.Sampler`, and a
``[sampler_name].yaml`` file, containing all possible user-specified options for the
sampler and their default values. Whatever option is defined in this file automatically
becomes an attribute of the sampler's instance.

To implement your own sampler, or an interface to an external one, simply create a folder
under the ``cobaya/cobaya/samplers/`` folder and include the two files described above.
Your class needs to inherit from the :class:`cobaya.Sampler` class below, and needs to
implement only the methods ``initialize``, ``run``, ``close``, and ``products``.

"""
# Python 2/3 compatibility
from __future__ import absolute_import, division, print_function

# Global
import os
import logging
import numpy as np

# Local
from cobaya.conventions import _sampler, _checkpoint_extension, _covmat_extension
from cobaya.conventions import _resume_default
from cobaya.tools import get_class
from cobaya.log import HandledException
from cobaya.yaml import yaml_load_file
from cobaya.mpi import am_single_or_primary_process

# Logger
log = logging.getLogger(__name__.split(".")[-1])


class Sampler(object):
    """Prototype of the sampler class."""

    # What you *must* implement to create your own sampler:

    def initialize(self):
        """
        Initializes the sampler: prepares the samples' collection,
        prepares the output, deals with MPI scheduling, imports an external sampler, etc.

        Options defined in the ``defaults.yaml`` file in the sampler's folder are
        automatically recognized as attributes, with the value given in the input file,
        if redefined there.

        The prior and likelihood are also accesible through the attributes with the same
        names.
        """
        pass

    def run(self):
        """
        Runs the main part of the algorithm of the sampler.
        Normally, it looks somewhat like

        .. code-block:: python

           while not [convergence criterion]:
               [do one more step]
               [update the collection of samples]
        """
        pass

    def close(self, exception_type, exception_value, traceback):
        """
        Finalizes the sampler, if something needs to be done
        (e.g. generating additional output).
        """
        pass

    def products(self):
        """
        Returns the products expected in a scripted call of cobaya,
        (e.g. a collection of samples or a list of them).
        """
        return None

    # Private methods: just ignore them:
    def __init__(self, info_sampler, model, output, resume=_resume_default, modules=None):
        """
        Actual initialization of the class. Loads the default and input information and
        call the custom ``initialize`` method.

        [Do not modify this one.]
        """
        self.name = self.__class__.__name__
        self.log = logging.getLogger(self.name)
        self.model = model
        self.output = output
        self.path_install = modules
        # Load info of the sampler
        for k in info_sampler:
            setattr(self, k, info_sampler[k])
        # Seed, if requested
        if getattr(self, "seed", None) is not None:
            self.log.warning("This run has been SEEDED with seed %d", self.seed)
            try:
                np.random.seed(self.seed)
            except TypeError:
                self.log.error("Seeds must be *integer*, but got %r with type %r",
                               self.seed, type(self.seed))
                raise HandledException
        # Load checkpoint info, if resuming
        self.resuming = resume
        if self.resuming:
            try:
                checkpoint_info = yaml_load_file(self.checkpoint_filename())
                try:
                    for k, v in checkpoint_info[_sampler][self.name].items():
                        setattr(self, k, v)
                    self.resuming = True
                    if am_single_or_primary_process():
                        self.log.info("Resuming from previous sample!")
                except KeyError:
                    if am_single_or_primary_process():
                        self.log.error("Checkpoint file found at '%s'"
                                       "but corresponds to a different sampler.",
                                       self.checkpoint_filename())
                        raise HandledException
            except (IOError, TypeError):
                pass
        else:
            try:
                os.remove(self.checkpoint_filename())
            except (OSError, TypeError):
                pass
        self.initialize()

    def checkpoint_filename(self):
        if self.output:
            return os.path.join(
                self.output.folder, self.output.prefix + _checkpoint_extension)
        return None

    def covmat_filename(self):
        if self.output:
            return os.path.join(
                self.output.folder, self.output.prefix + _covmat_extension)
        return None

    # Python magic for the "with" statement

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        # Reset seed
        if getattr(self, "seed", None) is not None:
            np.random.seed(self.seed)
        self.close(exception_type, exception_value, traceback)


def get_sampler(info_sampler, posterior, output_file,
                resume=_resume_default, modules=None):
    """
    Auxiliary function to retrieve and initialize the requested sampler.
    """
    if not info_sampler:
        log.error("No sampler given!")
        raise HandledException
    try:
        name = list(info_sampler)[0]
    except AttributeError:
        log.error("The sampler block must be a dictionary 'sampler: {options}'.")
        raise HandledException
    sampler_class = get_class(name, kind=_sampler)
    return sampler_class(
        info_sampler[name], posterior, output_file, resume=resume, modules=modules)



"""
.. module:: conventions

:Synopsis: Some physical constants and naming conventions
           to make the life of the maintainer easier.
:Author: Jesus Torrado

"""
# Package name (for importlib)
# (apparently __package__ is only defined if you import something locally.
_package = __name__.rpartition('.')[0]

# Names for block and fields in the input
_likelihood = "likelihood"
_prior = "prior"
_theory = "theory"
_sampler = "sampler"
_post = "post"
_params = "params"
_input_params = "input_params"
_output_params = "output_params"
_input_params_prefix = "input_params_prefix"
_output_params_prefix = "output_params_prefix"
_p_value = "value"
_p_dist = "dist"
_p_ref = "ref"
_p_label = "latex"
_p_renames = "renames"
_p_drop = "drop"
_p_derived = "derived"
_p_proposal = "proposal"
_debug = "debug"
_debug_default = False
_debug_file = "debug_file"
_output_prefix = "output"
_path_install = "modules"
_external = "external"
_resume = "resume"
_resume_default = False
_timing = "timing"
_force = "force"

# Separator for different names.
# Its manual inclusion in a string anywhere else (e.g. a parameter name) is forbidden.
_separator = "__"

# Names for the samples' fields internally and in the output
_weight = "weight"  # sample weight
_minuslogpost = "minuslogpost"  # log-posterior, or in general the total log-probability
_minuslogprior = "minuslogprior"  # log-prior
_chi2 = "chi2"  # chi^2 = -2 * loglik
_prior_1d_name = "0"

# Output files
_input_suffix = "input"
_full_suffix = "full"
_yaml_extensions = ".yaml", ".yml"
_checkpoint_extension = ".checkpoint"
_covmat_extension = ".covmat"

# Installation and container definitions
_modules_path_arg = "modules"
_modules_path = "/modules"
_code = "code"
_data = "data"
_covmats_file = "covmats_database.yaml"
_products_path = "/products"
_requirements_file = "requirements.yaml"
_help_file = "readme.md"

# Internal package structure
subfolders = {_likelihood: "likelihoods",
              _sampler: "samplers",
              _theory: "theories"}

# Approximate overhead of cobaya per posterior evaluation. Useful for blocking speeds
# After testing, it's mostly due to evaluating logpdf of scipy.stats 1d pdf's,
# in particular ~0.1ms per param
# (faster for most common cases, after manual override of logpdf)
_overhead_per_param = 5e-6

# Line width for console printing
_line_width = 120

# Physical constants
# ------------------
# Light speed
_c_km_s = 299792.458  # speed of light
_T_CMB_K = 2.72548  # CMB temperature
_h_J_s = 6.626070040e-34  # Planck's constant
_kB_J_K = 1.38064852e-23  # Boltzmann constant
